from __future__ import annotations

from dataclasses import dataclass

DEFAULT_MAX_LINES = 2000
DEFAULT_MAX_BYTES = 50 * 1024
GREP_MAX_LINE_LENGTH = 500


@dataclass(slots=True)
class TruncationResult:
    content: str
    truncated: bool
    truncated_by: str | None
    total_lines: int
    total_bytes: int
    output_lines: int
    output_bytes: int
    last_line_partial: bool
    first_line_exceeds_limit: bool
    max_lines: int
    max_bytes: int


def format_size(bytes_count: int) -> str:
    if bytes_count < 1024:
        return f"{bytes_count}B"
    if bytes_count < 1024 * 1024:
        return f"{bytes_count / 1024:.1f}KB"
    return f"{bytes_count / (1024 * 1024):.1f}MB"


def truncate_head(content: str, max_lines: int = DEFAULT_MAX_LINES, max_bytes: int = DEFAULT_MAX_BYTES) -> TruncationResult:
    total_bytes = len(content.encode("utf-8"))
    lines = content.split("\n")
    total_lines = len(lines)
    if total_lines <= max_lines and total_bytes <= max_bytes:
        return TruncationResult(
            content=content,
            truncated=False,
            truncated_by=None,
            total_lines=total_lines,
            total_bytes=total_bytes,
            output_lines=total_lines,
            output_bytes=total_bytes,
            last_line_partial=False,
            first_line_exceeds_limit=False,
            max_lines=max_lines,
            max_bytes=max_bytes,
        )

    if len(lines[0].encode("utf-8")) > max_bytes:
        return TruncationResult(
            content="",
            truncated=True,
            truncated_by="bytes",
            total_lines=total_lines,
            total_bytes=total_bytes,
            output_lines=0,
            output_bytes=0,
            last_line_partial=False,
            first_line_exceeds_limit=True,
            max_lines=max_lines,
            max_bytes=max_bytes,
        )

    out_lines: list[str] = []
    used = 0
    truncated_by = "lines"
    for index, line in enumerate(lines[:max_lines]):
        extra = len(line.encode("utf-8")) + (1 if index > 0 else 0)
        if used + extra > max_bytes:
            truncated_by = "bytes"
            break
        out_lines.append(line)
        used += extra

    output = "\n".join(out_lines)
    return TruncationResult(
        content=output,
        truncated=True,
        truncated_by=truncated_by,
        total_lines=total_lines,
        total_bytes=total_bytes,
        output_lines=len(out_lines),
        output_bytes=len(output.encode("utf-8")),
        last_line_partial=False,
        first_line_exceeds_limit=False,
        max_lines=max_lines,
        max_bytes=max_bytes,
    )


def _truncate_string_to_bytes_from_end(text: str, max_bytes: int) -> str:
    data = text.encode("utf-8")
    if len(data) <= max_bytes:
        return text
    chunk = data[-max_bytes:]
    while chunk and (chunk[0] & 0xC0) == 0x80:
        chunk = chunk[1:]
    return chunk.decode("utf-8", errors="ignore")


def truncate_tail(content: str, max_lines: int = DEFAULT_MAX_LINES, max_bytes: int = DEFAULT_MAX_BYTES) -> TruncationResult:
    total_bytes = len(content.encode("utf-8"))
    lines = content.split("\n")
    total_lines = len(lines)
    if total_lines <= max_lines and total_bytes <= max_bytes:
        return TruncationResult(
            content=content,
            truncated=False,
            truncated_by=None,
            total_lines=total_lines,
            total_bytes=total_bytes,
            output_lines=total_lines,
            output_bytes=total_bytes,
            last_line_partial=False,
            first_line_exceeds_limit=False,
            max_lines=max_lines,
            max_bytes=max_bytes,
        )

    out_lines: list[str] = []
    used = 0
    truncated_by = "lines"
    partial = False
    for index in range(len(lines) - 1, -1, -1):
        if len(out_lines) >= max_lines:
            break
        line = lines[index]
        extra = len(line.encode("utf-8")) + (1 if out_lines else 0)
        if used + extra > max_bytes:
            truncated_by = "bytes"
            if not out_lines:
                out_lines.insert(0, _truncate_string_to_bytes_from_end(line, max_bytes))
                used = len(out_lines[0].encode("utf-8"))
                partial = True
            break
        out_lines.insert(0, line)
        used += extra

    output = "\n".join(out_lines)
    return TruncationResult(
        content=output,
        truncated=True,
        truncated_by=truncated_by,
        total_lines=total_lines,
        total_bytes=total_bytes,
        output_lines=len(out_lines),
        output_bytes=len(output.encode("utf-8")),
        last_line_partial=partial,
        first_line_exceeds_limit=False,
        max_lines=max_lines,
        max_bytes=max_bytes,
    )


def truncate_line(line: str, max_chars: int = GREP_MAX_LINE_LENGTH) -> tuple[str, bool]:
    if len(line) <= max_chars:
        return line, False
    return f"{line[:max_chars]}... [truncated]", True



from __future__ import annotations

from .autocomplete import (
    AutocompleteItem,
    AutocompleteProvider,
    CombinedAutocompleteProvider,
    SlashCommand,
)
from .components import (
    Box,
    CancellableLoader,
    Editor,
    EditorOptions,
    EditorTheme,
    Image,
    ImageOptions,
    ImageTheme,
    Input,
    Loader,
    Markdown,
    SelectItem,
    SelectList,
    SelectListLayoutOptions,
    SelectListTheme,
    SelectListTruncatePrimaryContext,
    SettingItem,
    SettingsList,
    SettingsListTheme,
    Spacer,
    Text,
    TruncatedText,
)
from .editor_component import EditorComponent
from .fuzzy import FuzzyMatch, fuzzy_filter, fuzzy_match
from .keybindings import (
    DEFAULT_EDITOR_KEYBINDINGS,
    EditorAction,
    EditorKeybindingsManager,
    get_editor_keybindings,
    set_editor_keybindings,
)
from .keys import (
    Key,
    decode_kitty_printable,
    is_key_release,
    is_key_repeat,
    is_kitty_protocol_active,
    matches_key,
    parse_key,
    set_kitty_protocol_active,
)
from .kill_ring import KillRing
from .stdin_buffer import StdinBuffer
from .terminal import ProcessTerminal, Terminal
from .terminal_image import (
    CellDimensions,
    ImageDimensions,
    ImageRenderOptions,
    TerminalCapabilities,
    allocate_image_id,
    calculate_image_rows,
    delete_all_kitty_images,
    delete_kitty_image,
    detect_capabilities,
    encode_iterm2,
    encode_kitty,
    get_capabilities,
    get_cell_dimensions,
    get_gif_dimensions,
    get_image_dimensions,
    get_jpeg_dimensions,
    get_png_dimensions,
    get_webp_dimensions,
    image_fallback,
    is_image_line,
    render_image,
    reset_capabilities_cache,
    set_cell_dimensions,
)
from .tui import (
    CURSOR_MARKER,
    TUI,
    Component,
    Container,
    Focusable,
    OverlayAnchor,
    OverlayHandle,
    OverlayMargin,
    OverlayOptions,
    SizeValue,
    is_focusable,
)
from .undo_stack import UndoStack
from .utils import (
    apply_background_to_line,
    extract_ansi_code,
    extract_segments,
    get_segmenter,
    is_punctuation_char,
    is_whitespace_char,
    slice_by_column,
    slice_with_width,
    truncate_to_width,
    visible_width,
    wrap_text_with_ansi,
)

__all__ = [
    "AutocompleteItem",
    "AutocompleteProvider",
    "Box",
    "CancellableLoader",
    "CellDimensions",
    "CombinedAutocompleteProvider",
    "Component",
    "Container",
    "CURSOR_MARKER",
    "DEFAULT_EDITOR_KEYBINDINGS",
    "Editor",
    "EditorAction",
    "EditorComponent",
    "EditorOptions",
    "EditorTheme",
    "EditorKeybindingsManager",
    "Focusable",
    "FuzzyMatch",
    "Image",
    "ImageDimensions",
    "ImageOptions",
    "ImageRenderOptions",
    "ImageTheme",
    "Input",
    "Key",
    "KillRing",
    "Loader",
    "Markdown",
    "OverlayHandle",
    "OverlayMargin",
    "OverlayOptions",
    "OverlayAnchor",
    "ProcessTerminal",
    "SelectItem",
    "SelectList",
    "SelectListLayoutOptions",
    "SelectListTheme",
    "SelectListTruncatePrimaryContext",
    "SettingItem",
    "SettingsList",
    "SettingsListTheme",
    "SizeValue",
    "SlashCommand",
    "Spacer",
    "StdinBuffer",
    "Terminal",
    "TerminalCapabilities",
    "Text",
    "TruncatedText",
    "TUI",
    "UndoStack",
    "allocate_image_id",
    "apply_background_to_line",
    "calculate_image_rows",
    "decode_kitty_printable",
    "delete_all_kitty_images",
    "delete_kitty_image",
    "detect_capabilities",
    "encode_iterm2",
    "encode_kitty",
    "extract_ansi_code",
    "extract_segments",
    "fuzzy_filter",
    "fuzzy_match",
    "get_capabilities",
    "get_cell_dimensions",
    "get_gif_dimensions",
    "get_image_dimensions",
    "get_editor_keybindings",
    "get_jpeg_dimensions",
    "get_png_dimensions",
    "get_webp_dimensions",
    "get_segmenter",
    "image_fallback",
    "is_focusable",
    "is_image_line",
    "is_key_release",
    "is_key_repeat",
    "is_kitty_protocol_active",
    "is_punctuation_char",
    "is_whitespace_char",
    "matches_key",
    "parse_key",
    "render_image",
    "reset_capabilities_cache",
    "set_cell_dimensions",
    "set_editor_keybindings",
    "set_kitty_protocol_active",
    "slice_by_column",
    "slice_with_width",
    "truncate_to_width",
    "visible_width",
    "wrap_text_with_ansi",
]



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from ..agent import Agent
from ..agent.types import AgentMessage, ThinkingLevel
from ..ai import UserMessage
from ..ai.env_api_keys import get_env_default_model
from ..ai.models import get_model, supports_xhigh
from ..ai.types import ImageContent, TextContent
from .export_html import export_from_file
from .message_restore import restore_message
from .model_registry import ModelRegistry
from .package_manager import PackageManager
from .resource_loader import ResourceLoader
from .session_manager import SessionManager, infer_session_dir
from .settings_manager import SettingsManager
from .summarizer import estimate_tokens, generate_summary
from .types import CompactionPreparation, CompactionResult, PromptOptions, QueueMode, SessionInfo, SessionStateSnapshot

_THINKING_LEVELS: tuple[ThinkingLevel, ...] = ("off", "minimal", "low", "medium", "high", "xhigh")

CompactionHook = Callable[[CompactionPreparation], Awaitable[CompactionResult | None]]


def _estimate_message_entry_tokens(message: dict[str, Any]) -> int:
    content = message.get("content", "")
    if isinstance(content, str):
        text = content
    elif isinstance(content, list):
        text = "\n".join(block.get("text", "") for block in content if block.get("type") == "text")
    else:
        text = ""
    return max(len(text) // 4, 1) if text else 0


@dataclass(slots=True)
class AgentSessionConfig:
    agent: Agent
    session_manager: SessionManager
    settings_manager: SettingsManager
    cwd: str
    session_dir: str | None
    model_registry: ModelRegistry
    resource_loader: ResourceLoader


class AgentSession:
    def __init__(self, config: AgentSessionConfig) -> None:
        self.agent = config.agent
        self.session_manager = config.session_manager
        self.settings_manager = config.settings_manager
        self.cwd = config.cwd
        self.session_dir = str(Path(config.session_dir).resolve()) if config.session_dir is not None else None
        self.model_registry = config.model_registry
        self.resource_loader = config.resource_loader
        self._listeners: list[Callable[[Any], None]] = []
        self._compaction_hook: CompactionHook | None = None
        self.agent.subscribe(self._handle_agent_event)
        skills = self.resource_loader.get_skills().skills
        self.agent.set_system_prompt(
            self.resource_loader.build_system_prompt(self.agent.state.system_prompt, skills)
        )

    def subscribe(self, listener: Callable[[Any], None]) -> Callable[[], None]:
        self._listeners.append(listener)
        return lambda: self._listeners.remove(listener)

    def _emit(self, event: Any) -> None:
        for listener in list(self._listeners):
            listener(event)

    def _handle_agent_event(self, event: Any) -> None:
        if getattr(event, "type", None) == "message_end":
            self.session_manager.append_message(event.message)
        self._emit(event)

    def _restore_messages_from_current_branch(self) -> None:
        context = self.session_manager.build_runtime_context()
        self.agent.replace_messages([restore_message(message) for message in context["messages"] if isinstance(message, dict)])
        model = context.get("model")
        if model is not None:
            self.agent.set_model(self.model_registry.find(model["provider"], model["modelId"]))
        thinking_level = context.get("thinkingLevel")
        if thinking_level is not None:
            self.agent.set_thinking_level(thinking_level)

    @property
    def messages(self) -> list[AgentMessage]:
        return list(self.agent.state.messages)

    @property
    def session_file(self) -> str | None:
        return self.session_manager.get_session_file()

    @property
    def session_id(self) -> str:
        return self.session_manager.get_session_id()

    @property
    def model(self):
        return self.agent.state.model

    @property
    def thinking_level(self) -> ThinkingLevel:
        return self.agent.state.thinking_level

    @property
    def is_streaming(self) -> bool:
        return self.agent.state.is_streaming

    @property
    def system_prompt(self) -> str:
        return self.agent.state.system_prompt

    async def prompt(self, text: str, options: PromptOptions | None = None) -> None:
        opts = options or PromptOptions()
        images = opts.images
        if self.agent.state.is_streaming:
            if opts.streaming_behavior == "steer":
                await self.steer(text, images=images)
                return
            if opts.streaming_behavior == "followUp":
                await self.follow_up(text, images=images)
                return
            raise ValueError("Agent is already streaming. Use streaming_behavior='steer' or 'followUp'.")
        await self.agent.prompt(text, images=images)
        await self._maybe_auto_compact()

    async def steer(self, text: str, images: list[ImageContent] | None = None) -> None:
        content: str | list[Any] = text
        if images:
            content = [TextContent(text=text), *images]
        self.agent.steer(UserMessage(content=content))
        if not self.agent.state.is_streaming:
            await self.agent.continue_()
            await self._maybe_auto_compact()

    async def follow_up(self, text: str, images: list[ImageContent] | None = None) -> None:
        content: str | list[Any] = text
        if images:
            content = [TextContent(text=text), *images]
        self.agent.follow_up(UserMessage(content=content))
        if not self.agent.state.is_streaming:
            await self.agent.continue_()
            await self._maybe_auto_compact()

    async def abort(self) -> None:
        self.agent.abort()
        await self.agent.wait_for_idle()

    async def bash(self, command: str) -> dict[str, Any]:
        for tool in self.agent.state.tools:
            if tool.name == "bash":
                exit_code = 0
                try:
                    result = await tool.execute("bash", {"command": command}, asyncio.Event(), None)
                    output = "".join(
                        block.text for block in result.content if hasattr(block, "text")
                    )
                except ValueError as exc:
                    # bash tool raises ValueError with output on non-zero exit code
                    output = str(exc)
                    exit_code = 1
                message = {
                    "role": "bashExecution",
                    "command": command,
                    "output": output,
                    "exitCode": exit_code,
                    "cancelled": False,
                    "timestamp": 0,
                }
                self.session_manager.append_message(message)
                return message
        raise ValueError("bash tool is not enabled")

    async def new_session(self) -> None:
        self.agent.reset()
        current_file = self.session_manager.session_file
        if current_file is None:
            self.session_manager = SessionManager.in_memory(self.cwd)
            return

        session_root = self.session_dir or infer_session_dir(current_file)
        if session_root is None:
            self.session_manager = SessionManager.create(self.cwd)
            self.session_dir = infer_session_dir(self.session_manager.get_session_file() or "")
            return
        self.session_manager = SessionManager.create(self.cwd, session_root)

    def _apply_model(self, provider: str, model_id: str) -> None:
        model = self.model_registry.find(provider, model_id)
        self.agent.set_model(model)
        self.session_manager.append_model_change(provider, model_id)

    async def set_model(self, provider: str, model_id: str) -> None:
        self._apply_model(provider, model_id)

    async def cycle_model(self) -> Any:
        models = self.model_registry.list_models()
        if not models:
            raise ValueError("No models available")
        current_model = self.model
        if current_model is None:
            next_model = models[0]
        else:
            current_index = next(
                (
                    index
                    for index, model in enumerate(models)
                    if model.provider == current_model.provider and model.id == current_model.id
                ),
                -1,
            )
            next_model = models[(current_index + 1) % len(models)]
        await self.set_model(next_model.provider, next_model.id)
        return self.model

    def set_thinking_level(self, level: ThinkingLevel) -> None:
        self.agent.set_thinking_level(level)
        self.session_manager.append_thinking_level_change(level)

    def cycle_thinking_level(self) -> ThinkingLevel:
        levels = list(_THINKING_LEVELS)
        if not supports_xhigh(self.model):
            levels.remove("xhigh")
        current_level = self.thinking_level
        current_index = levels.index(current_level) if current_level in levels else -1
        next_level = levels[(current_index + 1) % len(levels)]
        self.set_thinking_level(next_level)
        return next_level

    def get_thinking_selector_schema(self) -> dict[str, Any]:
        items = [
            {
                "id": level,
                "value": level,
                "label": level,
                "description": "supported" if level != "xhigh" or supports_xhigh(self.model) else "unsupported",
                "isCurrent": level == self.thinking_level,
                "isSupported": level != "xhigh" or supports_xhigh(self.model),
                "position": index,
            }
            for index, level in enumerate(_THINKING_LEVELS)
            if level != "xhigh" or supports_xhigh(self.model)
        ]
        return {
            "currentThinkingLevel": self.thinking_level,
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    def set_steering_mode(self, mode: QueueMode) -> QueueMode:
        self.agent.set_steering_mode(mode)
        self.settings_manager.set_steering_mode(mode)
        return self.agent.get_steering_mode()

    def get_steering_selector_schema(self) -> dict[str, Any]:
        current_mode = self.agent.get_steering_mode()
        items = [
            {
                "id": mode,
                "value": mode,
                "label": mode,
                "description": "Queue every steering message" if mode == "all" else "Queue one steering message at a time",
                "isCurrent": mode == current_mode,
                "position": index,
            }
            for index, mode in enumerate(("one-at-a-time", "all"))
        ]
        return {
            "currentSteeringMode": current_mode,
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    def set_follow_up_mode(self, mode: QueueMode) -> QueueMode:
        self.agent.set_follow_up_mode(mode)
        self.settings_manager.set_follow_up_mode(mode)
        return self.agent.get_follow_up_mode()

    def get_follow_up_selector_schema(self) -> dict[str, Any]:
        current_mode = self.agent.get_follow_up_mode()
        items = [
            {
                "id": mode,
                "value": mode,
                "label": mode,
                "description": "Queue every follow-up message" if mode == "all" else "Queue one follow-up message at a time",
                "isCurrent": mode == current_mode,
                "position": index,
            }
            for index, mode in enumerate(("one-at-a-time", "all"))
        ]
        return {
            "currentFollowUpMode": current_mode,
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    def set_session_name(self, name: str | None) -> None:
        self.session_manager.set_session_name(name)

    def get_session_stats(self) -> dict[str, Any]:
        return self.session_manager.get_stats()

    def get_settings_snapshot(self) -> dict[str, Any]:
        settings = self.settings_manager.get_settings()
        return {
            "theme": self.get_theme(),
            "model": {"provider": self.model.provider, "id": self.model.id},
            "thinkingLevel": self.thinking_level,
            "steeringMode": self.agent.get_steering_mode(),
            "followUpMode": self.agent.get_follow_up_mode(),
            "quietStartup": settings.quiet_startup,
            "blockImages": settings.images.block_images,
            "showImages": settings.terminal.show_images,
            "enableSkillCommands": settings.enable_skill_commands,
            "transport": settings.transport,
            "retry": {
                "enabled": settings.retry.enabled,
                "maxRetries": settings.retry.max_retries,
                "baseDelayMs": settings.retry.base_delay_ms,
                "maxDelayMs": settings.retry.max_delay_ms,
            },
            "compaction": self.get_compaction_state(),
            "sessionName": self.session_manager.get_session_name(),
            "sessionFile": self.session_file,
        }

    def get_package_selector_schema(self) -> dict[str, Any]:
        manager = PackageManager(self.cwd, str(self.resource_loader.agent_dir), self.settings_manager)
        packages = manager.list()
        items = [
            {
                "id": package.source,
                "source": package.source,
                "scope": package.scope,
                "path": package.path,
                "label": package.source,
                "description": f"{package.scope}: {package.path}",
                "position": index,
            }
            for index, package in enumerate(packages)
        ]
        return {
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    def list_packages(self) -> list[dict[str, Any]]:
        manager = PackageManager(self.cwd, str(self.resource_loader.agent_dir), self.settings_manager)
        return [
            {
                "source": package.source,
                "scope": package.scope,
                "path": package.path,
            }
            for package in manager.list()
        ]

    def install_package(self, source: str, local: bool = False) -> dict[str, Any]:
        manager = PackageManager(self.cwd, str(self.resource_loader.agent_dir), self.settings_manager)
        path = manager.install(source, local=local)
        return {"source": source, "scope": "project" if local else "user", "path": path}

    def remove_package(self, source: str, local: bool = False) -> dict[str, Any]:
        manager = PackageManager(self.cwd, str(self.resource_loader.agent_dir), self.settings_manager)
        removed = manager.remove(source, local=local)
        return {"source": source, "scope": "project" if local else "user", "removed": removed}

    def update_packages(self, source: str | None = None) -> dict[str, Any]:
        manager = PackageManager(self.cwd, str(self.resource_loader.agent_dir), self.settings_manager)
        updated = manager.update(source)
        return {"updated": updated}

    def get_package_selector_item(self, source: str) -> dict[str, Any]:
        payload = self.get_selector_item("packages", source)
        payload["kind"] = "packages"
        return payload

    def get_resource_schema(self) -> dict[str, Any]:
        skills = self.resource_loader.get_skills()
        prompts = self.resource_loader.get_prompts()
        themes = self.resource_loader.get_themes()
        extensions = self.resource_loader.get_extensions()
        agents_files = self.resource_loader.get_agents_files()
        return {
            "counts": {
                "skills": len(skills.skills),
                "prompts": len(prompts.prompts),
                "themes": len(themes.themes),
                "extensions": len(extensions.extensions),
                "agentsFiles": len(agents_files.agents_files),
            },
            "skills": [
                {
                    "id": skill.name,
                    "name": skill.name,
                    "description": skill.description,
                    "source": skill.source,
                    "filePath": skill.file_path,
                    "position": index,
                }
                for index, skill in enumerate(skills.skills)
            ],
            "prompts": [
                {
                    "id": prompt.name,
                    "name": prompt.name,
                    "filePath": prompt.file_path,
                    "position": index,
                }
                for index, prompt in enumerate(prompts.prompts)
            ],
            "themes": [
                {
                    "id": theme.name,
                    "name": theme.name,
                    "filePath": theme.file_path,
                    "position": index,
                }
                for index, theme in enumerate(themes.themes)
            ],
            "extensions": [
                {
                    "id": getattr(module, "__name__", f"extension-{index}"),
                    "name": getattr(module, "__name__", f"extension-{index}"),
                    "position": index,
                }
                for index, module in enumerate(extensions.extensions)
            ],
            "agentsFiles": list(agents_files.agents_files),
            "diagnostics": {
                "skills": [{"type": item.type, "message": item.message} for item in skills.diagnostics],
                "prompts": [{"type": item.type, "message": item.message} for item in prompts.diagnostics],
                "themes": [{"type": item.type, "message": item.message} for item in themes.diagnostics],
                "extensions": list(extensions.errors),
            },
        }

    def get_resource_item(self, kind: str, item_id: str) -> dict[str, Any]:
        schema = self.get_resource_schema()
        normalized = kind.strip().lower().replace("-", "")
        if normalized == "agentsfiles":
            normalized = "agentsFiles"
        if normalized not in {"skills", "prompts", "themes", "extensions", "agentsFiles"}:
            raise ValueError(f"Unknown resource kind: {kind}")
        values = schema.get(normalized, [])
        if normalized == "agentsFiles":
            for item in values:
                if str(item) == item_id:
                    return {"kind": normalized, "requestedItemId": item_id, "resolvedItemId": str(item), "item": item}
        else:
            for item in values:
                if str(item.get("id")) == item_id or str(item.get("name")) == item_id:
                    resolved_item_id = str(item.get("id") or item.get("name") or item_id)
                    return {"kind": normalized, "requestedItemId": item_id, "resolvedItemId": resolved_item_id, "item": item}
        raise ValueError(f"Unknown resource item for {kind}: {item_id}")

    def get_command_schema(self) -> dict[str, Any]:
        commands = [
            {"name": "help", "group": "general", "order": 10, "description": "Show interactive help", "usage": "/help"},
            {"name": "session", "group": "session", "order": 20, "description": "Show current session stats", "usage": "/session"},
            {"name": "selector-item", "group": "resources", "order": 25, "description": "Show a single selector item", "usage": "/selector-item SELECTOR_ID ITEM_ID"},
            {"name": "settings", "group": "settings", "order": 30, "description": "Show or update settings", "usage": "/settings"},
            {"name": "settings-schema", "group": "settings", "order": 40, "description": "Show editable settings schema", "usage": "/settings schema"},
            {"name": "compaction", "group": "session", "order": 50, "description": "Show or update compaction settings", "usage": "/compaction"},
            {"name": "theme", "group": "appearance", "order": 60, "description": "Show or set theme", "usage": "/theme [NAME]"},
            {"name": "theme-schema", "group": "appearance", "order": 70, "description": "Show flat theme selector schema", "usage": "/theme schema"},
            {"name": "model", "group": "model", "order": 80, "description": "Show or set model", "usage": "/model [MODEL_ID]"},
            {"name": "model-schema", "group": "model", "order": 90, "description": "Show flat model selector schema", "usage": "/model schema"},
            {"name": "cycle-model", "group": "model", "order": 100, "description": "Cycle to the next model", "usage": "/cycle-model"},
            {"name": "thinking", "group": "model", "order": 110, "description": "Show, set, or cycle thinking level", "usage": "/thinking [LEVEL|cycle]"},
            {"name": "steering", "group": "interaction", "order": 120, "description": "Show or set steering queue mode", "usage": "/steering [MODE]"},
            {"name": "followup", "group": "interaction", "order": 130, "description": "Show or set follow-up queue mode", "usage": "/followup [MODE]"},
            {"name": "name", "group": "session", "order": 140, "description": "Set session name", "usage": "/name [TEXT]"},
            {"name": "new", "group": "session", "order": 150, "description": "Start a new session", "usage": "/new"},
            {"name": "reload", "group": "resources", "order": 160, "description": "Reload resources", "usage": "/reload"},
            {"name": "resources", "group": "resources", "order": 170, "description": "Show loaded resource counts", "usage": "/resources"},
            {"name": "resources-schema", "group": "resources", "order": 180, "description": "Show loaded resource details", "usage": "/resources schema"},
            {"name": "packages-schema", "group": "resources", "order": 190, "description": "Show installed package selector schema", "usage": "/packages schema"},
            {"name": "packages", "group": "resources", "order": 195, "description": "List installed packages", "usage": "/packages"},
            {"name": "packages-install", "group": "resources", "order": 196, "description": "Install a package from a local path or git URL", "usage": "/packages install SOURCE [local]"},
            {"name": "packages-remove", "group": "resources", "order": 197, "description": "Remove an installed package", "usage": "/packages remove SOURCE [local]"},
            {"name": "packages-update", "group": "resources", "order": 198, "description": "Update installed packages", "usage": "/packages update [SOURCE]"},
            {"name": "packages-uninstall", "group": "resources", "order": 199, "description": "Alias for packages remove", "usage": "/packages uninstall SOURCE [local]"},
            {"name": "packages-item", "group": "resources", "order": 200, "description": "Show details for an installed package", "usage": "/packages item SOURCE"},
            {"name": "resources-item", "group": "resources", "order": 201, "description": "Show details for a loaded resource", "usage": "/resources item KIND ID"},
            {"name": "resume", "group": "session", "order": 210, "description": "Resume most recent or matching session", "usage": "/resume [QUERY]"},
            {"name": "sessions", "group": "session", "order": 220, "description": "List saved sessions", "usage": "/sessions"},
            {"name": "sessions-schema", "group": "session", "order": 230, "description": "Show flat saved-session selector schema", "usage": "/sessions schema"},
            {"name": "switch", "group": "session", "order": 240, "description": "Switch to a session path, id, or name", "usage": "/switch QUERY"},
            {"name": "fork", "group": "tree", "order": 250, "description": "List or fork from user message", "usage": "/fork [ENTRY_ID]"},
            {"name": "tree", "group": "tree", "order": 260, "description": "Show or switch current session tree", "usage": "/tree [ENTRY_ID]"},
            {"name": "tree-schema", "group": "tree", "order": 270, "description": "Show flat session tree schema", "usage": "/tree schema"},
            {"name": "branch-summary", "group": "tree", "order": 280, "description": "Branch and append a summary entry", "usage": "/branch-summary ID TEXT"},
            {"name": "branch-summary-auto", "group": "tree", "order": 290, "description": "Generate and append a branch summary", "usage": "/branch-summary-auto ID [TEXT]"},
            {"name": "compact", "group": "session", "order": 300, "description": "Append a compaction summary entry", "usage": "/compact ID TOKENS TEXT"},
            {"name": "compact-auto", "group": "session", "order": 310, "description": "Generate and append a compaction summary", "usage": "/compact-auto ID [TEXT]"},
            {"name": "summarize", "group": "session", "order": 320, "description": "Generate a summary of the current session", "usage": "/summarize [TEXT]"},
            {"name": "last", "group": "session", "order": 330, "description": "Show last assistant text", "usage": "/last"},
            {"name": "quit", "group": "general", "order": 340, "description": "Exit interactive mode", "usage": "/quit"},
        ]
        groups = [
            {"id": "general", "label": "General", "order": 10},
            {"id": "session", "label": "Session", "order": 20},
            {"id": "settings", "label": "Settings", "order": 30},
            {"id": "appearance", "label": "Appearance", "order": 40},
            {"id": "model", "label": "Model", "order": 50},
            {"id": "interaction", "label": "Interaction", "order": 60},
            {"id": "resources", "label": "Resources", "order": 70},
            {"id": "tree", "label": "Tree", "order": 80},
        ]
        ordered_commands = sorted(commands, key=lambda item: int(item["order"]))
        return {
            "groups": groups,
            "itemOrder": [item["name"] for item in ordered_commands],
            "commands": ordered_commands,
        }

    def get_selector_registry(self) -> dict[str, Any]:
        selectors = [
            {
                "id": "commands",
                "label": "Commands",
                "kind": "list",
                "getter": "get_command_schema",
                "currentKey": None,
                "itemOrder": self.get_command_schema()["itemOrder"],
                "group": "general",
            },
            {
                "id": "settings",
                "label": "Settings",
                "kind": "fields",
                "getter": "get_settings_schema",
                "currentKey": None,
                "itemOrder": self.get_settings_schema()["fieldOrder"],
                "group": "settings",
            },
            {
                "id": "compaction",
                "label": "Compaction",
                "kind": "fields",
                "getter": "get_compaction_schema",
                "currentKey": "enabled",
                "itemOrder": self.get_compaction_schema()["fieldOrder"],
                "group": "settings",
            },
            {
                "id": "theme",
                "label": "Themes",
                "kind": "list",
                "getter": "get_theme_selector_schema",
                "currentKey": "currentTheme",
                "itemOrder": self.get_theme_selector_schema()["itemOrder"],
                "group": "appearance",
            },
            {
                "id": "model",
                "label": "Models",
                "kind": "list",
                "getter": "get_model_selector_schema",
                "currentKey": "currentModel",
                "itemOrder": self.get_model_selector_schema()["itemOrder"],
                "group": "model",
            },
            {
                "id": "thinking",
                "label": "Thinking",
                "kind": "list",
                "getter": "get_thinking_selector_schema",
                "currentKey": "currentThinkingLevel",
                "itemOrder": self.get_thinking_selector_schema()["itemOrder"],
                "group": "model",
            },
            {
                "id": "steering",
                "label": "Steering",
                "kind": "list",
                "getter": "get_steering_selector_schema",
                "currentKey": "currentSteeringMode",
                "itemOrder": self.get_steering_selector_schema()["itemOrder"],
                "group": "interaction",
            },
            {
                "id": "followup",
                "label": "Follow-up",
                "kind": "list",
                "getter": "get_follow_up_selector_schema",
                "currentKey": "currentFollowUpMode",
                "itemOrder": self.get_follow_up_selector_schema()["itemOrder"],
                "group": "interaction",
            },
            {
                "id": "sessions",
                "label": "Sessions",
                "kind": "list",
                "getter": "get_session_selector_schema",
                "currentKey": "currentSessionFile",
                "itemOrder": self.get_session_selector_schema()["itemOrder"],
                "group": "session",
            },
            {
                "id": "tree",
                "label": "Tree",
                "kind": "list",
                "getter": "get_tree_schema",
                "currentKey": "currentLeafId",
                "itemOrder": self.get_tree_schema()["itemOrder"],
                "group": "session",
            },
            {
                "id": "packages",
                "label": "Packages",
                "kind": "list",
                "getter": "get_package_selector_schema",
                "currentKey": None,
                "itemOrder": self.get_package_selector_schema()["itemOrder"],
                "group": "resources",
            },
            {
                "id": "resources",
                "label": "Resources",
                "kind": "object",
                "getter": "get_resource_schema",
                "currentKey": None,
                "itemOrder": ["skills", "prompts", "themes", "extensions", "agentsFiles"],
                "group": "resources",
            },
        ]
        groups = [
            {"id": "general", "label": "General", "order": 10},
            {"id": "settings", "label": "Settings", "order": 20},
            {"id": "appearance", "label": "Appearance", "order": 30},
            {"id": "model", "label": "Model", "order": 40},
            {"id": "interaction", "label": "Interaction", "order": 50},
            {"id": "session", "label": "Session", "order": 60},
            {"id": "resources", "label": "Resources", "order": 70},
        ]
        ordered = sorted(selectors, key=lambda item: (next(group["order"] for group in groups if group["id"] == item["group"]), item["label"]))
        return {
            "groups": groups,
            "itemOrder": [item["id"] for item in ordered],
            "selectors": ordered,
        }

    def get_selector(self, selector_id: str) -> dict[str, Any]:
        registry = self.get_selector_registry()
        selector = next((item for item in registry["selectors"] if item["id"] == selector_id), None)
        if selector is None:
            raise ValueError(f"Unknown selector: {selector_id}")
        getter = getattr(self, selector["getter"])
        payload = getter()
        return {
            "selector": selector,
            "preview": self._selector_preview(selector, payload),
            "data": payload,
        }

    def get_selector_item(self, selector_id: str, item_id: str) -> dict[str, Any]:
        payload = self.get_selector(selector_id)
        selector = payload["selector"]
        data = payload["data"]
        item = self._find_selector_item(data, item_id)
        if item is None:
            raise ValueError(f"Unknown item for selector {selector_id}: {item_id}")
        resolved_item_id = str(item.get("id") or item.get("name") or item_id) if isinstance(item, dict) else str(item)
        return {
            "selector": selector,
            "requestedItemId": item_id,
            "resolvedItemId": resolved_item_id,
            "item": item,
        }

    def _find_selector_item(self, payload: dict[str, Any], item_id: str) -> Any | None:
        kind_hint: str | None = None
        item_key = item_id
        known_kinds = {"skills", "prompts", "themes", "extensions", "agentsfiles"}
        for separator in (":", "/"):
            if separator in item_id:
                if separator == "/" and item_id.startswith("/"):
                    continue
                possible_kind, possible_item = item_id.split(separator, 1)
                possible_kind = possible_kind.strip().lower().replace("-", "")
                if possible_kind in known_kinds:
                    kind_hint = possible_kind
                    item_key = possible_item.strip()
                    break
        for key in ("items", "commands", "fields", "selectors", "groups", "skills", "prompts", "themes", "extensions", "agentsFiles"):
            values = payload.get(key)
            if isinstance(values, list):
                if kind_hint is not None:
                    normalized_key = key.lower().replace("-", "")
                    if normalized_key != kind_hint and not (key == "agentsFiles" and kind_hint == "agentsfiles"):
                        continue
                for value in values:
                    if key in {"extensions", "agentsFiles"} and str(value) == item_key:
                        return value
                    if isinstance(value, dict) and str(value.get("id")) == item_key:
                        return value
                    if isinstance(value, dict) and str(value.get("name")) == item_key:
                        return value
        return None

    def _selector_preview(self, selector: dict[str, Any], payload: dict[str, Any]) -> str:
        selector_id = selector["id"]
        if selector_id == "commands":
            return f'{len(payload.get("commands", []))} commands'
        if selector_id == "settings":
            return f'{len(payload.get("fields", []))} fields'
        if selector_id == "compaction":
            state = payload.get("state", {})
            return f'enabled={state.get("enabled")} reserve={state.get("reserveTokens")} keep={state.get("keepRecentTokens")}'
        if selector_id == "theme":
            return str(payload.get("currentTheme"))
        if selector_id == "model":
            current = payload.get("currentModel", {})
            return f'{current.get("provider")}/{current.get("id")}'
        if selector_id == "thinking":
            return str(payload.get("currentThinkingLevel"))
        if selector_id == "steering":
            return str(payload.get("currentSteeringMode"))
        if selector_id == "followup":
            return str(payload.get("currentFollowUpMode"))
        if selector_id == "sessions":
            return str(payload.get("currentSessionFile") or "none")
        if selector_id == "tree":
            return str(payload.get("currentLeafId") or "root")
        if selector_id == "packages":
            return f'{len(payload.get("items", []))} packages'
        if selector_id == "resources":
            counts = payload.get("counts", {})
            return (
                f"skills={counts.get('skills', 0)} "
                f"prompts={counts.get('prompts', 0)} "
                f"themes={counts.get('themes', 0)}"
            )
        return selector["label"]

    def get_settings_schema(self) -> dict[str, Any]:
        fields = [
                {
                    "id": "name",
                    "order": 10,
                    "label": "Session name",
                    "description": "Human-readable name for the current session",
                    "group": "session",
                    "type": "string",
                    "updatePath": "sessionName",
                    "command": "/settings name <text>",
                    "current": self.session_manager.get_session_name(),
                },
                {
                    "id": "model",
                    "order": 20,
                    "label": "Model",
                    "description": "Default model for new prompts in this session",
                    "group": "model",
                    "type": "select",
                    "updatePath": "model",
                    "command": "/settings model <model-id>",
                    "current": {"provider": self.model.provider, "id": self.model.id},
                    "options": [
                        {
                            "provider": model.provider,
                            "id": model.id,
                            "value": model.id,
                            "label": f"{model.provider}/{model.id}",
                        }
                        for model in self.get_available_models()
                    ],
                },
                {
                    "id": "theme",
                    "order": 30,
                    "label": "Theme",
                    "description": "Color theme for the interface",
                    "group": "appearance",
                    "type": "select",
                    "updatePath": "theme",
                    "command": "/settings theme <name>",
                    "current": self.get_theme(),
                    "options": [{"value": item["name"], "label": item["name"]} for item in self.get_themes()["themes"]],
                },
                {
                    "id": "thinking",
                    "order": 40,
                    "label": "Thinking level",
                    "description": "Reasoning intensity for the model",
                    "group": "model",
                    "type": "select",
                    "updatePath": "thinkingLevel",
                    "command": "/settings thinking <level>",
                    "current": self.thinking_level,
                    "options": [{"value": level, "label": level} for level in _THINKING_LEVELS if level != "xhigh" or supports_xhigh(self.model)],
                },
                {
                    "id": "steering",
                    "order": 50,
                    "label": "Steering mode",
                    "description": "How steering messages are queued while streaming",
                    "group": "interaction",
                    "type": "select",
                    "updatePath": "steeringMode",
                    "command": "/settings steering <mode>",
                    "current": self.agent.get_steering_mode(),
                    "options": [{"value": mode, "label": mode} for mode in ("one-at-a-time", "all")],
                },
                {
                    "id": "followup",
                    "order": 60,
                    "label": "Follow-up mode",
                    "description": "How follow-up messages are queued while streaming",
                    "group": "interaction",
                    "type": "select",
                    "updatePath": "followUpMode",
                    "command": "/settings followup <mode>",
                    "current": self.agent.get_follow_up_mode(),
                    "options": [{"value": mode, "label": mode} for mode in ("one-at-a-time", "all")],
                },
                {
                    "id": "quiet",
                    "order": 70,
                    "label": "Quiet startup",
                    "description": "Reduce startup output in interactive mode",
                    "group": "interaction",
                    "type": "boolean",
                    "updatePath": "quietStartup",
                    "command": "/settings quiet <true|false>",
                    "current": self.settings_manager.get_settings().quiet_startup,
                },
                {
                    "id": "block-images",
                    "order": 80,
                    "label": "Block images",
                    "description": "Prevent image inputs from being sent to the model",
                    "group": "appearance",
                    "type": "boolean",
                    "updatePath": "blockImages",
                    "command": "/settings block-images <true|false>",
                    "current": self.settings_manager.get_settings().images.block_images,
                },
                {
                    "id": "show-images",
                    "order": 90,
                    "label": "Show images",
                    "description": "Render images in supported terminals",
                    "group": "appearance",
                    "type": "boolean",
                    "updatePath": "showImages",
                    "command": "/settings show-images <true|false>",
                    "current": self.settings_manager.get_settings().terminal.show_images,
                },
                {
                    "id": "skill-commands",
                    "order": 100,
                    "label": "Skill commands",
                    "description": "Expose loaded skills as slash commands",
                    "group": "resources",
                    "type": "boolean",
                    "updatePath": "enableSkillCommands",
                    "command": "/settings skill-commands <true|false>",
                    "current": self.settings_manager.get_settings().enable_skill_commands,
                },
                {
                    "id": "transport",
                    "order": 110,
                    "label": "Transport",
                    "description": "Preferred provider transport",
                    "group": "network",
                    "type": "select",
                    "updatePath": "transport",
                    "command": "/settings transport <value>",
                    "current": self.settings_manager.get_settings().transport,
                    "options": [{"value": value, "label": value} for value in ("sse", "websocket", "auto")],
                },
                {
                    "id": "retry",
                    "order": 120,
                    "label": "Retry policy",
                    "description": "Automatic retry behavior for provider requests",
                    "group": "network",
                    "type": "object",
                    "updatePath": "retry",
                    "current": self.get_settings_snapshot()["retry"],
                    "fields": [
                        {"id": "enabled", "label": "Enabled", "type": "boolean", "updatePath": "retry.enabled", "command": "/settings retry enabled <true|false>"},
                        {"id": "maxRetries", "label": "Max retries", "type": "number", "updatePath": "retry.maxRetries", "command": "/settings retry max-retries <n>"},
                        {"id": "baseDelayMs", "label": "Base delay (ms)", "type": "number", "updatePath": "retry.baseDelayMs", "command": "/settings retry base-delay-ms <ms>"},
                        {"id": "maxDelayMs", "label": "Max delay (ms)", "type": "number", "updatePath": "retry.maxDelayMs", "command": "/settings retry max-delay-ms <ms>"},
                    ],
                },
                {
                    "id": "compaction",
                    "order": 130,
                    "label": "Compaction",
                    "description": "Conversation compaction thresholds and state",
                    "group": "session",
                    "type": "object",
                    "updatePath": "compaction",
                    "current": self.get_compaction_state(),
                    "fields": [
                        {"id": "enabled", "label": "Enabled", "type": "boolean", "updatePath": "compaction.enabled", "command": "/settings compaction enabled <true|false>"},
                        {"id": "reserveTokens", "label": "Reserve tokens", "type": "number", "updatePath": "compaction.reserveTokens", "command": "/settings compaction reserve <n>"},
                        {"id": "keepRecentTokens", "label": "Keep recent tokens", "type": "number", "updatePath": "compaction.keepRecentTokens", "command": "/settings compaction keep <n>"},
                        {"id": "estimatedTokens", "label": "Estimated tokens", "type": "number", "readonly": True},
                        {"id": "thresholdTokens", "label": "Threshold tokens", "type": "number", "readonly": True},
                        {"id": "shouldCompact", "label": "Should compact", "type": "boolean", "readonly": True},
                        {"id": "contextWindow", "label": "Context window", "type": "number", "readonly": True},
                    ],
                },
            ]
        groups = [
            {"id": "session", "label": "Session", "order": 10},
            {"id": "model", "label": "Model", "order": 20},
            {"id": "interaction", "label": "Interaction", "order": 30},
            {"id": "appearance", "label": "Appearance", "order": 40},
            {"id": "resources", "label": "Resources", "order": 50},
            {"id": "network", "label": "Network", "order": 60},
        ]
        return {
            "groups": sorted(groups, key=lambda item: int(item.get("order", 0))),
            "fieldOrder": [field["id"] for field in sorted(fields, key=lambda item: int(item.get("order", 0)))],
            "fields": fields,
        }

    def update_settings(self, patch: dict[str, Any]) -> dict[str, Any]:
        if "model" in patch:
            model_patch = patch["model"]
            if isinstance(model_patch, dict):
                provider = str(model_patch.get("provider") or self.model.provider)
                model_id = str(model_patch.get("id") or self.model.id)
            else:
                provider = self.model.provider
                model_id = str(model_patch)
            self._apply_model(provider, model_id)
        if "theme" in patch:
            self.set_theme(str(patch["theme"]))
        if "sessionName" in patch:
            self.set_session_name(None if patch["sessionName"] is None else str(patch["sessionName"]))
        if "quietStartup" in patch:
            self.settings_manager.set_quiet_startup(bool(patch["quietStartup"]))
        if "blockImages" in patch:
            self.settings_manager.set_block_images(bool(patch["blockImages"]))
        if "showImages" in patch:
            self.settings_manager.set_show_images(bool(patch["showImages"]))
        if "enableSkillCommands" in patch:
            self.settings_manager.set_enable_skill_commands(bool(patch["enableSkillCommands"]))
        if "transport" in patch:
            self.settings_manager.set_transport(str(patch["transport"]))
        if "thinkingLevel" in patch:
            self.set_thinking_level(patch["thinkingLevel"])
        if "steeringMode" in patch:
            self.set_steering_mode(patch["steeringMode"])
        if "followUpMode" in patch:
            self.set_follow_up_mode(patch["followUpMode"])
        retry = patch.get("retry")
        if isinstance(retry, dict):
            if "enabled" in retry:
                self.settings_manager.set_retry_enabled(bool(retry["enabled"]))
            if "maxRetries" in retry:
                self.settings_manager.set_retry_max_retries(int(retry["maxRetries"]))
            if "baseDelayMs" in retry:
                self.settings_manager.set_retry_base_delay_ms(int(retry["baseDelayMs"]))
            if "maxDelayMs" in retry:
                self.settings_manager.set_retry_max_delay_ms(int(retry["maxDelayMs"]))
        compaction = patch.get("compaction")
        if isinstance(compaction, dict):
            if "enabled" in compaction:
                self.set_compaction_enabled(bool(compaction["enabled"]))
            if "reserveTokens" in compaction:
                self.set_compaction_reserve_tokens(int(compaction["reserveTokens"]))
            if "keepRecentTokens" in compaction:
                self.set_compaction_keep_recent_tokens(int(compaction["keepRecentTokens"]))
        return self.get_settings_snapshot()

    def get_theme(self) -> str:
        return self.settings_manager.get_theme()

    def get_themes(self) -> dict[str, Any]:
        loaded = self.resource_loader.get_themes()
        return {
            "currentTheme": self.get_theme(),
            "themes": [{"name": theme.name, "filePath": theme.file_path} for theme in loaded.themes],
            "diagnostics": [{"type": item.type, "message": item.message} for item in loaded.diagnostics],
        }

    def get_theme_selector_schema(self) -> dict[str, Any]:
        themes = self.get_themes()
        current_theme = themes["currentTheme"]
        items = [
            {
                "id": item["name"],
                "value": item["name"],
                "label": item["name"],
                "description": item["filePath"],
                "isCurrent": item["name"] == current_theme,
                "position": index,
            }
            for index, item in enumerate(themes["themes"])
        ]
        return {
            "currentTheme": current_theme,
            "itemOrder": [item["id"] for item in items],
            "items": items,
            "diagnostics": themes["diagnostics"],
        }

    def set_theme(self, theme: str) -> dict[str, Any]:
        available = {item["name"] for item in self.get_themes()["themes"]}
        if available and theme not in available:
            raise ValueError(f"Unknown theme: {theme}")
        self.settings_manager.set_theme(theme)
        return self.get_themes()

    def get_compaction_state(self) -> dict[str, Any]:
        compaction = self.settings_manager.get_settings().compaction
        estimated_tokens = estimate_tokens(self.messages)
        threshold = max(self.model.context_window - compaction.reserve_tokens, 0)
        return {
            "enabled": compaction.enabled,
            "estimatedTokens": estimated_tokens,
            "thresholdTokens": threshold,
            "reserveTokens": compaction.reserve_tokens,
            "keepRecentTokens": compaction.keep_recent_tokens,
            "shouldCompact": compaction.enabled and estimated_tokens > threshold,
            "contextWindow": self.model.context_window,
        }

    def get_compaction_schema(self) -> dict[str, Any]:
        state = self.get_compaction_state()
        fields = [
            {
                "id": "enabled",
                "label": "Enabled",
                "type": "boolean",
                "current": state["enabled"],
                "updatePath": "compaction.enabled",
                "command": "/compaction enabled <true|false>",
                "readonly": False,
                "order": 10,
            },
            {
                "id": "reserveTokens",
                "label": "Reserve tokens",
                "type": "number",
                "current": state["reserveTokens"],
                "updatePath": "compaction.reserveTokens",
                "command": "/compaction reserve <n>",
                "readonly": False,
                "order": 20,
            },
            {
                "id": "keepRecentTokens",
                "label": "Keep recent tokens",
                "type": "number",
                "current": state["keepRecentTokens"],
                "updatePath": "compaction.keepRecentTokens",
                "command": "/compaction keep <n>",
                "readonly": False,
                "order": 30,
            },
            {
                "id": "estimatedTokens",
                "label": "Estimated tokens",
                "type": "number",
                "current": state["estimatedTokens"],
                "readonly": True,
                "order": 40,
            },
            {
                "id": "thresholdTokens",
                "label": "Threshold tokens",
                "type": "number",
                "current": state["thresholdTokens"],
                "readonly": True,
                "order": 50,
            },
            {
                "id": "shouldCompact",
                "label": "Should compact",
                "type": "boolean",
                "current": state["shouldCompact"],
                "readonly": True,
                "order": 60,
            },
            {
                "id": "contextWindow",
                "label": "Context window",
                "type": "number",
                "current": state["contextWindow"],
                "readonly": True,
                "order": 70,
            },
        ]
        return {
            "fieldOrder": [field["id"] for field in sorted(fields, key=lambda item: int(item["order"]))],
            "fields": fields,
            "state": state,
        }

    def set_compaction_enabled(self, enabled: bool) -> dict[str, Any]:
        self.settings_manager.set_compaction_enabled(enabled)
        return self.get_compaction_state()

    def set_compaction_reserve_tokens(self, reserve_tokens: int) -> dict[str, Any]:
        self.settings_manager.set_compaction_reserve_tokens(reserve_tokens)
        return self.get_compaction_state()

    def set_compaction_keep_recent_tokens(self, keep_recent_tokens: int) -> dict[str, Any]:
        self.settings_manager.set_compaction_keep_recent_tokens(keep_recent_tokens)
        return self.get_compaction_state()

    def set_compaction_hook(self, hook: CompactionHook | None) -> None:
        """Register a hook called before compaction summarization.

        The hook receives a :class:`CompactionPreparation` and should return
        a :class:`CompactionResult` to override the default summarization, or
        ``None`` to fall back to the built-in behaviour.
        """
        self._compaction_hook = hook

    def get_available_models(self) -> list[Any]:
        return self.model_registry.list_models()

    def get_model_selector_schema(self) -> dict[str, Any]:
        models = self.get_available_models()
        current_model = {"provider": self.model.provider, "id": self.model.id}
        items = [
            {
                "id": f"{model.provider}/{model.id}",
                "provider": model.provider,
                "modelId": model.id,
                "label": f"{model.provider}/{model.id}",
                "description": f"contextWindow={model.context_window}",
                "isCurrent": model.provider == current_model["provider"] and model.id == current_model["id"],
                "position": index,
            }
            for index, model in enumerate(models)
        ]
        return {
            "currentModel": current_model,
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    async def export_to_html(self, output_path: str | None = None) -> str:
        session_file = self.session_file
        if session_file is None:
            raise ValueError("Cannot export in-memory session to HTML")
        return export_from_file(session_file, output_path)

    async def switch_session(self, session_path: str) -> bool:
        manager = SessionManager.open(session_path)
        self.session_manager = manager
        inferred_session_dir = infer_session_dir(session_path)
        if inferred_session_dir is not None:
            self.session_dir = inferred_session_dir
        self.agent.reset()
        self._restore_messages_from_current_branch()
        return True

    def resolve_session(self, query: str) -> str:
        current_file = self.session_file
        if current_file is not None:
            current_info = SessionManager.read_session_info(current_file)
            normalized_query = query.strip().lower()
            current_path = Path(current_file)
            if normalized_query in {
                current_path.name.lower(),
                current_path.stem.lower(),
                self.session_id.lower(),
            }:
                return current_file
            if current_info is not None:
                if current_info.name and normalized_query in current_info.name.lower():
                    return current_file
                if normalized_query and normalized_query in current_info.first_message.lower():
                    return current_file
        return str(SessionManager.resolve_session(self.cwd, query, self.session_dir))

    async def resolve_and_switch_session(self, query: str) -> str:
        session_path = self.resolve_session(query)
        await self.switch_session(session_path)
        return session_path

    def get_fork_messages(self) -> list[dict[str, str]]:
        return self.session_manager.get_user_messages_for_forking()

    async def fork(self, entry_id: str) -> dict[str, Any]:
        new_manager, selected_text = self.session_manager.fork_to_new_manager(entry_id, self.session_dir)
        self.session_manager = new_manager
        self._restore_messages_from_current_branch()
        return {"selectedText": selected_text, "cancelled": False}

    def branch_with_summary(self, entry_id: str | None, summary: str) -> dict[str, Any]:
        summary = summary.strip()
        if not summary:
            raise ValueError("Summary cannot be empty")
        summary_entry_id = self.session_manager.branch_with_summary(entry_id, summary)
        self._restore_messages_from_current_branch()
        return {
            "entryId": entry_id,
            "summaryEntryId": summary_entry_id,
            "leafId": self.session_manager.get_leaf_id(),
            "cancelled": False,
        }

    def compact(self, first_kept_entry_id: str, summary: str, tokens_before: int) -> dict[str, Any]:
        summary = summary.strip()
        if not summary:
            raise ValueError("Summary cannot be empty")
        compaction_entry_id = self.session_manager.append_compaction(summary, first_kept_entry_id, tokens_before)
        self._restore_messages_from_current_branch()
        return {
            "compactionEntryId": compaction_entry_id,
            "leafId": self.session_manager.get_leaf_id(),
            "cancelled": False,
        }

    async def generate_summary(self, custom_instructions: str | None = None) -> dict[str, Any]:
        summary = await generate_summary(self.model, self.messages, custom_instructions)
        return {"summary": summary}

    async def auto_branch_with_summary(self, entry_id: str | None, custom_instructions: str | None = None) -> dict[str, Any]:
        summary = await generate_summary(self.model, self.messages, custom_instructions)
        return self.branch_with_summary(entry_id, summary)

    async def auto_compact(self, first_kept_entry_id: str, custom_instructions: str | None = None) -> dict[str, Any]:
        tokens_before = estimate_tokens(self.messages)

        # Build preparation and try hook first.
        preparation = CompactionPreparation(
            first_kept_entry_id=first_kept_entry_id,
            messages_to_summarize=list(self.messages),
            tokens_before=tokens_before,
            previous_summary=None,
            custom_instructions=custom_instructions,
        )

        hook_result: CompactionResult | None = None
        if self._compaction_hook is not None:
            hook_result = await self._compaction_hook(preparation)

        if hook_result is not None:
            compaction_entry_id = self.session_manager.append_compaction(
                hook_result.summary,
                hook_result.first_kept_entry_id,
                hook_result.tokens_before,
                details=hook_result.details,
                from_hook=True,
            )
            self._restore_messages_from_current_branch()
            return {
                "compactionEntryId": compaction_entry_id,
                "leafId": self.session_manager.get_leaf_id(),
                "cancelled": False,
            }

        # Fall back to default summarization.
        summary = await generate_summary(self.model, self.messages, custom_instructions)
        return self.compact(first_kept_entry_id, summary, tokens_before)

    async def _maybe_auto_compact(self) -> None:
        settings = self.settings_manager.get_settings().compaction
        if not settings.enabled:
            return
        current_messages = self.messages
        estimated_tokens = estimate_tokens(current_messages)
        threshold = max(self.model.context_window - settings.reserve_tokens, 0)
        if estimated_tokens <= threshold:
            return

        first_kept_entry_id = self._find_first_kept_entry_id(settings.keep_recent_tokens)
        if first_kept_entry_id is None:
            return

        result = await self.auto_compact(first_kept_entry_id)
        self._emit({"type": "auto_compaction_end", "data": result})

    def _find_first_kept_entry_id(self, keep_recent_tokens: int) -> str | None:
        branch_entries = self.session_manager.get_branch_entries()
        message_entries = [entry for entry in branch_entries if entry["type"] == "message"]
        if len(message_entries) < 2:
            return None

        kept: list[dict[str, Any]] = []
        kept_tokens = 0
        for entry in reversed(message_entries):
            kept.insert(0, entry)
            kept_tokens += _estimate_message_entry_tokens(entry["message"])
            if kept_tokens >= keep_recent_tokens:
                break

        if not kept or kept[0]["id"] == message_entries[0]["id"]:
            return None
        return kept[0]["id"]

    def get_last_assistant_text(self) -> str | None:
        return self.session_manager.get_last_assistant_text()

    def find_most_recent_session(self) -> str | None:
        path = SessionManager.find_most_recent_session(self.cwd, self.session_dir)
        return None if path is None else str(path)

    def list_sessions(self) -> list[str]:
        return [str(path) for path in SessionManager.list_sessions(self.cwd, self.session_dir)]

    def list_session_infos(self) -> list[SessionInfo]:
        return SessionManager.list_session_infos(self.cwd, self.session_dir)

    def get_session_selector_schema(self) -> dict[str, Any]:
        current_file = self.session_file
        infos = self.list_session_infos()
        items: list[dict[str, Any]] = []
        seen_paths: set[str] = set()

        for index, info in enumerate(infos):
            seen_paths.add(info.path)
            items.append(
                {
                    "id": info.id,
                    "path": info.path,
                    "name": info.name,
                    "label": info.name or Path(info.path).stem,
                    "description": info.first_message,
                    "messageCount": info.message_count,
                    "created": info.created,
                    "modified": info.modified,
                    "firstMessage": info.first_message,
                    "cwd": info.cwd,
                    "isCurrent": info.path == current_file,
                    "position": index,
                }
            )

        if current_file is not None and current_file not in seen_paths:
            current_info = SessionManager.read_session_info(current_file)
            if current_info is not None:
                items.insert(
                    0,
                    {
                        "id": current_info.id,
                        "path": current_info.path,
                        "name": current_info.name,
                        "label": current_info.name or Path(current_info.path).stem,
                        "description": current_info.first_message,
                        "messageCount": current_info.message_count,
                        "created": current_info.created,
                        "modified": current_info.modified,
                        "firstMessage": current_info.first_message,
                        "cwd": current_info.cwd,
                        "isCurrent": True,
                        "position": 0,
                    },
                )
                for index, item in enumerate(items):
                    item["position"] = index

        return {
            "currentSessionFile": current_file,
            "itemOrder": [item["path"] for item in items],
            "items": items,
        }

    def get_tree(self) -> list[dict[str, Any]]:
        return self.session_manager.get_tree()

    def get_tree_schema(self) -> dict[str, Any]:
        branch_ids = {entry["id"] for entry in self.session_manager.get_branch_entries()}
        items: list[dict[str, Any]] = []

        def visit(nodes: list[dict[str, Any]], depth: int) -> None:
            for index, node in enumerate(nodes):
                entry = node["entry"]
                children = node["children"]
                items.append(
                    {
                        "id": entry["id"],
                        "parentId": entry.get("parentId"),
                        "type": entry["type"],
                        "label": entry["label"],
                        "timestamp": entry["timestamp"],
                        "depth": depth,
                        "childCount": len(children),
                        "isLeaf": bool(node.get("isLeaf")),
                        "isOnCurrentBranch": entry["id"] in branch_ids,
                        "position": index,
                    }
                )
                visit(children, depth + 1)

        tree = self.get_tree()
        visit(tree, 0)
        return {
            "currentLeafId": self.session_manager.get_leaf_id(),
            "itemOrder": [item["id"] for item in items],
            "items": items,
        }

    def branch(self, entry_id: str | None) -> dict[str, Any]:
        self.session_manager.branch(entry_id)
        self._restore_messages_from_current_branch()
        return {"entryId": entry_id, "leafId": self.session_manager.get_leaf_id(), "cancelled": False}

    def get_state(self) -> SessionStateSnapshot:
        return SessionStateSnapshot(
            model=self.agent.state.model,
            thinking_level=self.agent.state.thinking_level,
            is_streaming=self.agent.state.is_streaming,
            session_file=self.session_file,
            session_id=self.session_id,
            message_count=len(self.agent.state.messages),
            steering_mode=self.agent.get_steering_mode(),
            follow_up_mode=self.agent.get_follow_up_mode(),
            pending_message_count=self.agent.get_pending_message_count(),
        )


def create_default_model():
    return get_model("openai", get_env_default_model() or "gpt-4o-mini")



from __future__ import annotations

import asyncio
from collections.abc import Mapping, Sequence
from dataclasses import asdict, is_dataclass
from typing import Any, Callable, Literal

from ..ai.stream import stream as default_stream
from ..ai.types import ImageContent, Model, SimpleStreamOptions, TextContent, UserMessage
from .agent_loop import (
    AgentEndEvent,
    AgentLoopConfig,
    AgentMessage,
    MessageEndEvent,
    MessageStartEvent,
    MessageUpdateEvent,
    ToolExecutionEndEvent,
    ToolExecutionStartEvent,
    TurnEndEvent,
    _run_agent_loop,
    default_convert_to_llm,
    normalize_simple_stream_options,
)
from .types import AgentContext, AgentOptions, AgentState, AgentTool, ThinkingLevel


def _normalize_agent_options(options: AgentOptions | Mapping[str, Any] | None) -> AgentOptions:
    if options is None:
        return AgentOptions()
    if isinstance(options, AgentOptions):
        return options
    if is_dataclass(options):
        return AgentOptions(**asdict(options))
    if isinstance(options, Mapping):
        return AgentOptions(**dict(options))
    raise TypeError(f"Unsupported agent options type: {type(options)!r}")


def _build_state(initial_state: Mapping[str, Any] | None) -> AgentState:
    state = AgentState()
    if not initial_state:
        return state

    for key, value in initial_state.items():
        if hasattr(state, key):
            setattr(state, key, value)
    return state


class Agent:
    def __init__(self, options: AgentOptions | Mapping[str, Any] | None = None) -> None:
        opts = _normalize_agent_options(options)
        self._state = _build_state(opts.initial_state)
        self.convert_to_llm = opts.convert_to_llm or default_convert_to_llm
        self.transform_context = opts.transform_context
        self.steering_mode = opts.steering_mode
        self.follow_up_mode = opts.follow_up_mode
        self.stream_fn = opts.stream_fn or default_stream
        self.get_api_key = opts.get_api_key
        self.stream_options = opts.stream_options
        self.tool_execution = opts.tool_execution
        self._before_tool_call = opts.before_tool_call
        self._after_tool_call = opts.after_tool_call
        self._listeners: set[Callable[[Any], None]] = set()
        self._running_future: asyncio.Future[None] | None = None
        self._current_task: asyncio.Task[Any] | None = None
        self._cancel_event: asyncio.Event | None = None
        self._steering_queue: list[AgentMessage] = []
        self._follow_up_queue: list[AgentMessage] = []

    @property
    def state(self) -> AgentState:
        return self._state

    def subscribe(self, fn: Callable[[Any], None]) -> Callable[[], None]:
        self._listeners.add(fn)
        return lambda: self._listeners.discard(fn)

    def set_system_prompt(self, value: str) -> None:
        self._state.system_prompt = value

    def set_model(self, model: Model) -> None:
        self._state.model = model

    def set_thinking_level(self, level: ThinkingLevel) -> None:
        self._state.thinking_level = level

    def set_steering_mode(self, mode: Literal["all", "one-at-a-time"]) -> None:
        self.steering_mode = mode

    def get_steering_mode(self) -> Literal["all", "one-at-a-time"]:
        return self.steering_mode

    def set_follow_up_mode(self, mode: Literal["all", "one-at-a-time"]) -> None:
        self.follow_up_mode = mode

    def get_follow_up_mode(self) -> Literal["all", "one-at-a-time"]:
        return self.follow_up_mode

    def set_tools(self, tools: list[AgentTool[Any, Any]]) -> None:
        self._state.tools = tools

    def replace_messages(self, messages: list[AgentMessage]) -> None:
        self._state.messages = list(messages)

    def append_message(self, message: AgentMessage) -> None:
        self._state.messages = [*self._state.messages, message]

    def clear_messages(self) -> None:
        self._state.messages = []

    def steer(self, message: AgentMessage) -> None:
        self._steering_queue.append(message)

    def follow_up(self, message: AgentMessage) -> None:
        self._follow_up_queue.append(message)

    def clear_steering_queue(self) -> None:
        self._steering_queue = []

    def clear_follow_up_queue(self) -> None:
        self._follow_up_queue = []

    def clear_all_queues(self) -> None:
        self._steering_queue = []
        self._follow_up_queue = []

    def has_queued_messages(self) -> bool:
        return bool(self._steering_queue or self._follow_up_queue)

    def get_pending_message_count(self) -> int:
        return len(self._steering_queue) + len(self._follow_up_queue)

    def abort(self) -> None:
        if self._cancel_event is not None:
            self._cancel_event.set()

    async def wait_for_idle(self) -> None:
        if self._running_future is not None:
            await self._running_future

    def reset(self) -> None:
        self._state.messages = []
        self._state.is_streaming = False
        self._state.stream_message = None
        self._state.pending_tool_calls = set()
        self._state.error = None
        self.clear_all_queues()

    async def prompt(
        self,
        input: str | AgentMessage | Sequence[AgentMessage],
        images: list[ImageContent] | None = None,
    ) -> None:
        """Start a new turn from text or agent messages.

        Custom AgentMessageLike inputs are accepted into agent state as-is. They
        only reach the model if convert_to_llm maps them to standard ai.Message
        objects before the request is sent.
        """

        if self._state.is_streaming:
            raise ValueError(
                "Agent is already processing a prompt. Use steer() or follow_up() to queue messages, or wait for completion.",
            )

        messages = self._normalize_prompt_input(input, images)
        await self._run_with_messages(messages)

    async def continue_(self) -> None:
        if self._state.is_streaming:
            raise ValueError("Agent is already processing. Wait for completion before continuing.")

        if not self._state.messages:
            raise ValueError("No messages to continue from")

        if _agent_message_role(self._state.messages[-1]) == "assistant":
            queued_steering = self._dequeue_steering_messages()
            if queued_steering:
                await self._run_with_messages(queued_steering, skip_initial_steering_poll=True)
                return

            queued_follow_up = self._dequeue_follow_up_messages()
            if queued_follow_up:
                await self._run_with_messages(queued_follow_up)
                return

            context_messages = _strip_trailing_assistant_messages(self._state.messages)
            if not context_messages:
                raise ValueError("Cannot continue from message role: assistant")
            await self._run_with_messages(None, context_messages=context_messages)
            return

        await self._run_with_messages(None, context_messages=list(self._state.messages))

    def _process_loop_event(self, event: Any) -> None:
        if isinstance(event, MessageStartEvent):
            self._state.stream_message = event.message
        elif isinstance(event, MessageUpdateEvent):
            self._state.stream_message = event.message
        elif isinstance(event, MessageEndEvent):
            self._state.stream_message = None
            self.append_message(event.message)
            if _agent_message_role(event.message) == "assistant" and _assistant_error_message(event.message):
                self._state.error = _assistant_error_message(event.message)
        elif isinstance(event, ToolExecutionStartEvent):
            self._state.pending_tool_calls = {*self._state.pending_tool_calls, event.tool_call_id}
        elif isinstance(event, ToolExecutionEndEvent):
            self._state.pending_tool_calls = {
                tool_call_id for tool_call_id in self._state.pending_tool_calls if tool_call_id != event.tool_call_id
            }
        elif isinstance(event, TurnEndEvent):
            if _agent_message_role(event.message) == "assistant" and _assistant_error_message(event.message):
                self._state.error = _assistant_error_message(event.message)
        elif isinstance(event, AgentEndEvent):
            self._state.is_streaming = False
            self._state.stream_message = None

        for listener in list(self._listeners):
            listener(event)

    def _normalize_prompt_input(
        self,
        input: str | AgentMessage | Sequence[AgentMessage],
        images: list[ImageContent] | None = None,
    ) -> list[AgentMessage]:
        # String inputs are normalized here. Custom app messages are left
        # untouched so convert_to_llm can handle them in full context.
        if isinstance(input, str):
            content: list[TextContent | ImageContent] = [TextContent(text=input)]
            if images:
                content.extend(images)
            return [
                UserMessage(
                    content=input if not images else content,
                )
            ]

        if isinstance(input, Sequence) and not hasattr(input, "role"):
            return list(input)

        return [input]

    def _dequeue_steering_messages(self) -> list[AgentMessage]:
        if self.steering_mode == "one-at-a-time":
            if not self._steering_queue:
                return []
            first = self._steering_queue[0]
            self._steering_queue = self._steering_queue[1:]
            return [first]

        steering = list(self._steering_queue)
        self._steering_queue = []
        return steering

    def _dequeue_follow_up_messages(self) -> list[AgentMessage]:
        if self.follow_up_mode == "one-at-a-time":
            if not self._follow_up_queue:
                return []
            first = self._follow_up_queue[0]
            self._follow_up_queue = self._follow_up_queue[1:]
            return [first]

        follow_up = list(self._follow_up_queue)
        self._follow_up_queue = []
        return follow_up

    def _build_stream_options(self) -> SimpleStreamOptions | dict[str, Any] | None:
        options = normalize_simple_stream_options(self.stream_options)
        if self._state.thinking_level != "off":
            options.reasoning = self._state.thinking_level
        return options

    def _build_loop_config(self, skip_initial_steering_poll: bool = False) -> AgentLoopConfig:
        options = self._build_stream_options()

        async def get_steering_messages() -> list[AgentMessage]:
            nonlocal skip_initial_steering_poll
            if skip_initial_steering_poll:
                skip_initial_steering_poll = False
                return []
            return self._dequeue_steering_messages()

        async def get_follow_up_messages() -> list[AgentMessage]:
            return self._dequeue_follow_up_messages()

        return AgentLoopConfig(
            model=self._state.model,
            convert_to_llm=self.convert_to_llm,
            transform_context=self.transform_context,
            get_api_key=self.get_api_key,
            get_steering_messages=get_steering_messages,
            get_follow_up_messages=get_follow_up_messages,
            tool_execution=self.tool_execution,
            before_tool_call=self._before_tool_call,
            after_tool_call=self._after_tool_call,
            stream_options=options,
        )

    async def _run_with_messages(
        self,
        messages: list[AgentMessage] | None,
        *,
        skip_initial_steering_poll: bool = False,
        context_messages: list[AgentMessage] | None = None,
    ) -> None:
        self._running_future = asyncio.get_running_loop().create_future()
        self._current_task = asyncio.current_task()
        self._cancel_event = asyncio.Event()
        self._state.is_streaming = True
        self._state.stream_message = None
        self._state.error = None

        loop_config = self._build_loop_config(skip_initial_steering_poll=skip_initial_steering_poll)
        context = AgentContext(
            system_prompt=self._state.system_prompt,
            messages=list(self._state.messages if context_messages is None else context_messages),
            tools=self._state.tools,
        )

        try:
            await _run_agent_loop(
                messages or [],
                context,
                loop_config,
                self._process_loop_event,
                self._cancel_event,
                self.stream_fn,
            )
        finally:
            self._state.is_streaming = False
            self._state.stream_message = None
            self._state.pending_tool_calls = set()
            self._current_task = None
            self._cancel_event = None
            if self._running_future is not None and not self._running_future.done():
                self._running_future.set_result(None)
            self._running_future = None


def _agent_message_role(message: AgentMessage) -> str | None:
    return getattr(message, "role", None)


def _assistant_error_message(message: AgentMessage) -> str | None:
    return getattr(message, "error_message", None)


def _strip_trailing_assistant_messages(messages: Sequence[AgentMessage]) -> list[AgentMessage]:
    trimmed = list(messages)
    while trimmed and _agent_message_role(trimmed[-1]) == "assistant":
        trimmed.pop()
    return trimmed



from __future__ import annotations

import os
from pathlib import Path

APP_NAME = "pi"
CONFIG_DIR_NAME = ".pi"
ENV_AGENT_DIR = "PI_CODING_AGENT_DIR"
VERSION = "0.1.0"


def get_agent_dir() -> Path:
    override = os.getenv(ENV_AGENT_DIR)
    if override:
        return Path(override).expanduser().resolve()
    return (Path.home() / CONFIG_DIR_NAME / "agent").resolve()


def get_settings_path() -> Path:
    return get_agent_dir() / "settings.json"


def get_auth_path() -> Path:
    return get_agent_dir() / "auth.json"


def get_models_path() -> Path:
    return get_agent_dir() / "models.json"


def get_prompts_dir() -> Path:
    return get_agent_dir() / "prompts"


def get_sessions_dir() -> Path:
    return get_agent_dir() / "sessions"


def get_extensions_dir() -> Path:
    return get_agent_dir() / "extensions"


def get_skills_dir() -> Path:
    return get_agent_dir() / "skills"


def get_themes_dir() -> Path:
    return get_agent_dir() / "themes"


def get_project_pi_dir(cwd: str | Path) -> Path:
    return Path(cwd).resolve() / CONFIG_DIR_NAME


def get_project_settings_path(cwd: str | Path) -> Path:
    return get_project_pi_dir(cwd) / "settings.json"


def get_project_extensions_dir(cwd: str | Path) -> Path:
    return get_project_pi_dir(cwd) / "extensions"


def get_project_skills_dir(cwd: str | Path) -> Path:
    return get_project_pi_dir(cwd) / "skills"


def get_project_prompts_dir(cwd: str | Path) -> Path:
    return get_project_pi_dir(cwd) / "prompts"


def get_project_themes_dir(cwd: str | Path) -> Path:
    return get_project_pi_dir(cwd) / "themes"


def get_project_agents_path(cwd: str | Path) -> Path:
    return Path(cwd).resolve() / "AGENTS.md"


def encode_cwd_for_session_dir(cwd: str | Path) -> str:
    path = str(Path(cwd).resolve())
    return f"--{path.replace('/', '-').replace(':', '')}--"



from __future__ import annotations

import json
from pathlib import Path
from typing import Any

from .config import get_project_settings_path, get_settings_path
from .types import (
    CompactionSettings,
    ImageSettings,
    RetrySettings,
    Settings,
    TerminalSettings,
)


def _deep_merge(base: dict[str, Any], override: dict[str, Any]) -> dict[str, Any]:
    merged = dict(base)
    for key, value in override.items():
        if isinstance(value, dict) and isinstance(merged.get(key), dict):
            merged[key] = _deep_merge(merged[key], value)
        else:
            merged[key] = value
    return merged


class SettingsManager:
    def __init__(self, cwd: str, agent_dir: Path) -> None:
        self.cwd = cwd
        self.agent_dir = agent_dir
        self.global_path = get_settings_path() if agent_dir == get_settings_path().parent else agent_dir / "settings.json"
        self.project_path = get_project_settings_path(cwd)
        self._project_overrides: dict[str, Any] = self._load_json(self.project_path)
        self._settings = self._load()

    @classmethod
    def create(cls, cwd: str, agent_dir: str | Path) -> "SettingsManager":
        return cls(cwd, Path(agent_dir))

    def _load_json(self, path: Path) -> dict[str, Any]:
        if not path.exists():
            return {}
        text = path.read_text(encoding="utf-8").strip()
        if not text:
            return {}
        try:
            return json.loads(text)
        except json.JSONDecodeError:
            return {}

    def _from_dict(self, data: dict[str, Any]) -> Settings:
        retry_data = data.get("retry", {})
        compaction_data = data.get("compaction", {})
        images = ImageSettings(
            auto_resize=data.get("images", {}).get("autoResize", True),
            block_images=data.get("images", {}).get("blockImages", False),
        )
        terminal = TerminalSettings(
            show_images=data.get("terminal", {}).get("showImages", True),
            clear_on_shrink=data.get("terminal", {}).get("clearOnShrink", False),
        )
        retry = RetrySettings(
            enabled=retry_data.get("enabled", True),
            max_retries=retry_data.get("maxRetries", retry_data.get("max_retries", 3)),
            base_delay_ms=retry_data.get("baseDelayMs", retry_data.get("base_delay_ms", 2000)),
            max_delay_ms=retry_data.get("maxDelayMs", retry_data.get("max_delay_ms", 60_000)),
        )
        compaction = CompactionSettings(
            enabled=compaction_data.get("enabled", True),
            reserve_tokens=compaction_data.get("reserveTokens", compaction_data.get("reserve_tokens", 16_384)),
            keep_recent_tokens=compaction_data.get("keepRecentTokens", compaction_data.get("keep_recent_tokens", 20_000)),
        )
        return Settings(
            default_provider=data.get("defaultProvider"),
            default_model=data.get("defaultModel"),
            default_thinking_level=data.get("defaultThinkingLevel"),
            theme=data.get("theme", "dark"),
            quiet_startup=data.get("quietStartup", False),
            steering_mode=data.get("steeringMode", "one-at-a-time"),
            follow_up_mode=data.get("followUpMode", "one-at-a-time"),
            transport=data.get("transport", "sse"),
            shell_command_prefix=data.get("shellCommandPrefix"),
            shell_path=data.get("shellPath"),
            enabled_models=list(data.get("enabledModels", [])),
            packages=list(data.get("packages", [])),
            extensions=list(data.get("extensions", [])),
            skills=list(data.get("skills", [])),
            prompts=list(data.get("prompts", [])),
            themes=list(data.get("themes", [])),
            enable_skill_commands=data.get("enableSkillCommands", True),
            retry=retry,
            compaction=compaction,
            images=images,
            terminal=terminal,
        )

    def _load(self) -> Settings:
        merged = _deep_merge(self._load_json(self.global_path), self._load_json(self.project_path))
        return self._from_dict(merged)

    def reload(self) -> None:
        self._settings = self._load()

    def _to_dict(self) -> dict[str, Any]:
        return {
            "defaultProvider": self._settings.default_provider,
            "defaultModel": self._settings.default_model,
            "defaultThinkingLevel": self._settings.default_thinking_level,
            "theme": self._settings.theme,
            "quietStartup": self._settings.quiet_startup,
            "steeringMode": self._settings.steering_mode,
            "followUpMode": self._settings.follow_up_mode,
            "transport": self._settings.transport,
            "shellCommandPrefix": self._settings.shell_command_prefix,
            "shellPath": self._settings.shell_path,
            "enabledModels": self._settings.enabled_models,
            "packages": self._settings.packages,
            "extensions": self._settings.extensions,
            "skills": self._settings.skills,
            "prompts": self._settings.prompts,
            "themes": self._settings.themes,
            "enableSkillCommands": self._settings.enable_skill_commands,
            "retry": {
                "enabled": self._settings.retry.enabled,
                "maxRetries": self._settings.retry.max_retries,
                "baseDelayMs": self._settings.retry.base_delay_ms,
                "maxDelayMs": self._settings.retry.max_delay_ms,
            },
            "compaction": {
                "enabled": self._settings.compaction.enabled,
                "reserveTokens": self._settings.compaction.reserve_tokens,
                "keepRecentTokens": self._settings.compaction.keep_recent_tokens,
            },
            "images": {
                "autoResize": self._settings.images.auto_resize,
                "blockImages": self._settings.images.block_images,
            },
            "terminal": {
                "showImages": self._settings.terminal.show_images,
                "clearOnShrink": self._settings.terminal.clear_on_shrink,
            },
        }

    def save_project_settings(self) -> None:
        self.project_path.parent.mkdir(parents=True, exist_ok=True)
        self.project_path.write_text(json.dumps(self._project_overrides, indent=2), encoding="utf-8")

    def get_settings(self) -> Settings:
        return self._settings

    def get_default_provider(self) -> str | None:
        return self._settings.default_provider

    def get_default_model(self) -> str | None:
        return self._settings.default_model

    def get_default_thinking_level(self):
        return self._settings.default_thinking_level

    def get_theme(self) -> str:
        return self._settings.theme

    def set_theme(self, theme: str) -> None:
        self._settings.theme = theme
        self._project_overrides["theme"] = theme
        self.save_project_settings()

    def get_quiet_startup(self) -> bool:
        return self._settings.quiet_startup

    def set_quiet_startup(self, enabled: bool) -> None:
        self._settings.quiet_startup = enabled
        self._project_overrides["quietStartup"] = enabled
        self.save_project_settings()

    def get_steering_mode(self):
        return self._settings.steering_mode

    def set_steering_mode(self, mode: str) -> None:
        self._settings.steering_mode = mode
        self._project_overrides["steeringMode"] = mode
        self.save_project_settings()

    def get_follow_up_mode(self):
        return self._settings.follow_up_mode

    def set_follow_up_mode(self, mode: str) -> None:
        self._settings.follow_up_mode = mode
        self._project_overrides["followUpMode"] = mode
        self.save_project_settings()

    def get_transport(self):
        return self._settings.transport

    def set_transport(self, transport: str) -> None:
        self._settings.transport = transport
        self._project_overrides["transport"] = transport
        self.save_project_settings()

    def get_shell_command_prefix(self) -> str | None:
        return self._settings.shell_command_prefix

    def get_block_images(self) -> bool:
        return self._settings.images.block_images

    def set_block_images(self, enabled: bool) -> None:
        self._settings.images.block_images = enabled
        self._project_overrides.setdefault("images", {})["blockImages"] = enabled
        self.save_project_settings()

    def set_show_images(self, enabled: bool) -> None:
        self._settings.terminal.show_images = enabled
        self._project_overrides.setdefault("terminal", {})["showImages"] = enabled
        self.save_project_settings()

    def set_enable_skill_commands(self, enabled: bool) -> None:
        self._settings.enable_skill_commands = enabled
        self._project_overrides["enableSkillCommands"] = enabled
        self.save_project_settings()

    def set_retry_enabled(self, enabled: bool) -> None:
        self._settings.retry.enabled = enabled
        self._project_overrides.setdefault("retry", {})["enabled"] = enabled
        self.save_project_settings()

    def set_retry_max_retries(self, max_retries: int) -> None:
        self._settings.retry.max_retries = max_retries
        self._project_overrides.setdefault("retry", {})["maxRetries"] = max_retries
        self.save_project_settings()

    def set_retry_base_delay_ms(self, base_delay_ms: int) -> None:
        self._settings.retry.base_delay_ms = base_delay_ms
        self._project_overrides.setdefault("retry", {})["baseDelayMs"] = base_delay_ms
        self.save_project_settings()

    def set_retry_max_delay_ms(self, max_delay_ms: int) -> None:
        self._settings.retry.max_delay_ms = max_delay_ms
        self._project_overrides.setdefault("retry", {})["maxDelayMs"] = max_delay_ms
        self.save_project_settings()

    def set_compaction_enabled(self, enabled: bool) -> None:
        self._settings.compaction.enabled = enabled
        self._project_overrides.setdefault("compaction", {})["enabled"] = enabled
        self.save_project_settings()

    def set_compaction_reserve_tokens(self, reserve_tokens: int) -> None:
        self._settings.compaction.reserve_tokens = reserve_tokens
        self._project_overrides.setdefault("compaction", {})["reserveTokens"] = reserve_tokens
        self.save_project_settings()

    def set_compaction_keep_recent_tokens(self, keep_recent_tokens: int) -> None:
        self._settings.compaction.keep_recent_tokens = keep_recent_tokens
        self._project_overrides.setdefault("compaction", {})["keepRecentTokens"] = keep_recent_tokens
        self.save_project_settings()



from __future__ import annotations

import json
import subprocess
from dataclasses import dataclass
from pathlib import Path

from kitpaw.pi_agent.tui import (
    Input,
    Markdown,
    OverlayOptions,
    SelectItem,
    SelectList,
    SettingItem,
    SettingsList,
    Text,
)

ROOT = Path(__file__).resolve().parent.parent
NODE = "node"
TSX_LOADER = ROOT / "ref/pi-mono/node_modules/tsx/dist/loader.mjs"
SNAPSHOT_SCRIPT = ROOT / "tests/tui_parity/snapshot.ts"
REFERENCE_SCRIPT = ROOT / "tests/tui_parity/reference.ts"


SELECT_THEME = {
    "selected_prefix": lambda text: text,
    "selected_text": lambda text: text,
    "description": lambda text: text,
    "scroll_info": lambda text: text,
    "no_match": lambda text: text,
}

MARKDOWN_THEME = {
    "heading": lambda text: text,
    "link": lambda text: text,
    "link_url": lambda text: text,
    "code": lambda text: text,
    "code_block": lambda text: text,
    "code_block_border": lambda text: text,
    "quote": lambda text: text,
    "quote_border": lambda text: text,
    "hr": lambda text: text,
    "list_bullet": lambda text: text,
    "bold": lambda text: text,
    "italic": lambda text: text,
    "strikethrough": lambda text: text,
    "underline": lambda text: text,
}

STYLED_MARKDOWN_THEME = {
    "heading": lambda text: text,
    "link": lambda text: text,
    "link_url": lambda text: f"\x1b[2m{text}\x1b[0m",
    "code": lambda text: f"\x1b[33m{text}\x1b[0m",
    "code_block": lambda text: text,
    "code_block_border": lambda text: text,
    "quote": lambda text: f"\x1b[3m{text}\x1b[0m",
    "quote_border": lambda text: text,
    "hr": lambda text: text,
    "list_bullet": lambda text: text,
    "bold": lambda text: f"\x1b[1m{text}\x1b[0m",
    "italic": lambda text: f"\x1b[3m{text}\x1b[0m",
    "strikethrough": lambda text: text,
    "underline": lambda text: f"\x1b[4m{text}\x1b[0m",
}

SETTINGS_THEME = {
    "label": lambda text, selected: text,
    "value": lambda text, selected: text,
    "description": lambda text: text,
    "cursor": "→ ",
    "hint": lambda text: text,
}


@dataclass
class RecordingTerminal:
    columns: int = 20
    rows: int = 6
    writes: list[str] = None  # type: ignore[assignment]

    def __post_init__(self) -> None:
        self.writes = []

    def start(self, on_input, on_resize) -> None:  # noqa: ANN001
        self.on_input = on_input
        self.on_resize = on_resize

    def stop(self) -> None:
        return None

    async def drain_input(self, max_ms: int = 1000, idle_ms: int = 50) -> None:  # noqa: ARG002
        return None

    def write(self, data: str) -> None:
        self.writes.append(data)

    @property
    def kitty_protocol_active(self) -> bool:
        return True

    def move_by(self, lines: int) -> None:
        if lines > 0:
            self.write(f"\x1b[{lines}B")
        elif lines < 0:
            self.write(f"\x1b[{-lines}A")

    def hide_cursor(self) -> None:
        self.write("\x1b[?25l")

    def show_cursor(self) -> None:
        self.write("\x1b[?25h")

    def clear_line(self) -> None:
        self.write("\x1b[K")

    def clear_from_cursor(self) -> None:
        self.write("\x1b[J")

    def clear_screen(self) -> None:
        self.write("\x1b[2J\x1b[H")

    def set_title(self, title: str) -> None:
        self.write(f"\x1b]0;{title}\x07")


def run_node_snapshot(writes: list[str], columns: int = 20, rows: int = 6) -> dict[str, object]:
    payload = {"writes": writes, "columns": columns, "rows": rows}
    result = subprocess.run(
        [NODE, "--import", str(TSX_LOADER), str(SNAPSHOT_SCRIPT)],
        input=json.dumps(payload),
        text=True,
        capture_output=True,
        check=True,
        cwd=str(ROOT),
    )
    return json.loads(result.stdout)


def run_node_reference(scenario: str) -> dict[str, object]:
    result = subprocess.run(
        [NODE, "--import", str(TSX_LOADER), str(REFERENCE_SCRIPT), scenario],
        text=True,
        capture_output=True,
        check=True,
        cwd=str(ROOT),
    )
    return json.loads(result.stdout)


def render_python_text() -> dict[str, object]:
    terminal = RecordingTerminal()
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("hello"))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_input() -> dict[str, object]:
    terminal = RecordingTerminal()
    from kitpaw.pi_agent.tui import TUI

    input_widget = Input()
    input_widget.focused = True
    input_widget.set_value("hello")
    tui = TUI(terminal)
    tui.add_child(input_widget)
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_select_list() -> dict[str, object]:
    terminal = RecordingTerminal(columns=80)
    from kitpaw.pi_agent.tui import TUI

    list_widget = SelectList(
        [
            SelectItem(value="short", label="short", description="short description"),
            SelectItem(
                value="very-long-command-name-that-needs-truncation",
                label="very-long-command-name-that-needs-truncation",
                description="long description",
            ),
        ],
        5,
        SELECT_THEME,
    )
    tui = TUI(terminal)
    tui.add_child(list_widget)
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("Line 1\nLine 2\nLine 3"))
    tui.show_overlay(Text("OVERLAY"), OverlayOptions(anchor="center", width=10))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_top_left() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("base"))
    tui.show_overlay(Text("TOP-LEFT"), OverlayOptions(anchor="top-left", width=10))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_bottom_right() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("base"))
    tui.show_overlay(Text("BTM-RIGHT"), OverlayOptions(anchor="bottom-right", width=10))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_visible_rule() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("base"))
    tui.show_overlay(
        Text("HIDDEN"),
        OverlayOptions(anchor="center", width=10, visible=lambda columns, _rows: columns >= 30),
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_width_percent_min() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("base"))
    tui.show_overlay(
        Text("XXXXXXXXXXXXXXX"),
        OverlayOptions(anchor="center", width="50%", min_width=12),
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_short_content() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(Text("Line 1\nLine 2\nLine 3"))
    tui.show_overlay(Text("OVERLAY_TOP\nOVERLAY_MID\nOVERLAY_BOT"))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_overlay_style_reset() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    component = _MutableComponent(["\x1b[3mXXXXXXXXXXXXXXXXXXXX\x1b[23m", "INPUT"])
    tui = TUI(terminal)
    tui.add_child(component)
    tui.show_overlay(Text("OVR"), OverlayOptions(row=0, col=5, width=3))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_list() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "- Item 1\n  - Nested 1.1\n  - Nested 1.2\n- Item 2",
            0,
            0,
            MARKDOWN_THEME,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_explicit_link() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "[click here](https://example.com)",
            0,
            0,
            MARKDOWN_THEME,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_blockquote_wrap() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "> This is a very long blockquote line that should wrap to multiple lines when rendered",
            0,
            0,
            MARKDOWN_THEME,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_table() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "| Name | Value |\n| --- | --- |\n| Foo | Bar |\n| Baz | Qux |",
            0,
            0,
            MARKDOWN_THEME,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_code_fence() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "```py\nprint('hello')\nprint('world')\n```",
            0,
            0,
            MARKDOWN_THEME,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_markdown_style_reset() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        Markdown(
            "This is thinking with `inline code`",
            1,
            0,
            STYLED_MARKDOWN_THEME,
            {
                "color": lambda text: f"\x1b[90m{text}\x1b[0m",
                "italic": True,
            },
        )
    )
    tui.add_child(Text("INPUT"))
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_settings_list() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    tui = TUI(terminal)
    tui.add_child(
        SettingsList(
            [
                SettingItem(id="theme", label="Theme", current_value="light", description="UI theme", values=["light", "dark"]),
                SettingItem(id="model", label="Model", current_value="gpt-5"),
            ],
            5,
            SETTINGS_THEME,
            lambda _id, _value: None,
            lambda: None,
        )
    )
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


class _ParitySubmenu:
    def __init__(self, current_value: str) -> None:
        self.current_value = current_value

    def render(self, width: int) -> list[str]:  # noqa: ARG002
        return [f"submenu:{self.current_value}"]

    def handle_input(self, data: str) -> None:  # noqa: ARG002
        return None

    def invalidate(self) -> None:
        return None


def render_python_settings_list_submenu_open() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    settings = SettingsList(
        [
            SettingItem(
                id="theme",
                label="Theme",
                current_value="light",
                submenu=lambda current_value, done: _ParitySubmenu(current_value),
            )
        ],
        5,
        SETTINGS_THEME,
        lambda _id, _value: None,
        lambda: None,
    )
    settings.handle_input("\r")

    tui = TUI(terminal)
    tui.add_child(settings)
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


class _MutableComponent:
    def __init__(self, lines: list[str]) -> None:
        self.lines = lines

    def render(self, _width: int) -> list[str]:
        return list(self.lines)

    def handle_input(self, _data: str) -> None:
        return None

    def invalidate(self) -> None:
        return None


def render_python_tui_middle_line_change() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    component = _MutableComponent(["Header", "Working...", "Footer"])
    tui = TUI(terminal)
    tui.add_child(component)
    tui.start()
    component.lines = ["Header", "Working /", "Footer"]
    tui.request_render()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_tui_clear_then_content() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    component = _MutableComponent(["Line 0", "Line 1", "Line 2"])
    tui = TUI(terminal)
    tui.add_child(component)
    tui.start()
    component.lines = []
    tui.request_render()
    component.lines = ["New Line 0", "New Line 1"]
    tui.request_render()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def render_python_tui_style_reset() -> dict[str, object]:
    terminal = RecordingTerminal(columns=20, rows=6)
    from kitpaw.pi_agent.tui import TUI

    component = _MutableComponent(["\x1b[3mItalic", "Plain"])
    tui = TUI(terminal)
    tui.add_child(component)
    tui.start()
    return run_node_snapshot(terminal.writes, terminal.columns, terminal.rows)


def test_parity_text_basic() -> None:
    python_snapshot = render_python_text()
    ts_snapshot = run_node_reference("text-basic")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]
    assert python_snapshot["cursor"] == ts_snapshot["cursor"]


def test_parity_input_focus() -> None:
    python_snapshot = render_python_input()
    ts_snapshot = run_node_reference("input-focus")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]
    assert python_snapshot["cursor"] == ts_snapshot["cursor"]


def test_parity_select_list_basic() -> None:
    python_snapshot = render_python_select_list()
    ts_snapshot = run_node_reference("select-list-basic")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_centered() -> None:
    python_snapshot = render_python_overlay()
    ts_snapshot = run_node_reference("overlay-centered")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_top_left() -> None:
    python_snapshot = render_python_overlay_top_left()
    ts_snapshot = run_node_reference("overlay-top-left")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_bottom_right() -> None:
    python_snapshot = render_python_overlay_bottom_right()
    ts_snapshot = run_node_reference("overlay-bottom-right")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_visible_rule() -> None:
    python_snapshot = render_python_overlay_visible_rule()
    ts_snapshot = run_node_reference("overlay-visible-rule")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_width_percent_min() -> None:
    python_snapshot = render_python_overlay_width_percent_min()
    ts_snapshot = run_node_reference("overlay-width-percent-min")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_short_content() -> None:
    python_snapshot = render_python_overlay_short_content()
    ts_snapshot = run_node_reference("overlay-short-content")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_overlay_style_reset() -> None:
    python_snapshot = render_python_overlay_style_reset()
    ts_snapshot = run_node_reference("overlay-style-reset")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]
    assert python_snapshot["italic"] == ts_snapshot["italic"]


def test_parity_markdown_list() -> None:
    python_snapshot = render_python_markdown_list()
    ts_snapshot = run_node_reference("markdown-list")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_markdown_explicit_link() -> None:
    python_snapshot = render_python_markdown_explicit_link()
    ts_snapshot = run_node_reference("markdown-explicit-link")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_markdown_blockquote_wrap() -> None:
    python_snapshot = render_python_markdown_blockquote_wrap()
    ts_snapshot = run_node_reference("markdown-blockquote-wrap")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_markdown_table() -> None:
    python_snapshot = render_python_markdown_table()
    ts_snapshot = run_node_reference("markdown-table")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_markdown_code_fence() -> None:
    python_snapshot = render_python_markdown_code_fence()
    ts_snapshot = run_node_reference("markdown-code-fence")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_markdown_style_reset() -> None:
    python_snapshot = render_python_markdown_style_reset()
    ts_snapshot = run_node_reference("markdown-style-reset")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]
    assert python_snapshot["italic"] == ts_snapshot["italic"]


def test_parity_settings_list_basic() -> None:
    python_snapshot = render_python_settings_list()
    ts_snapshot = run_node_reference("settings-list-basic")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_settings_list_submenu_open() -> None:
    python_snapshot = render_python_settings_list_submenu_open()
    ts_snapshot = run_node_reference("settings-list-submenu-open")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_tui_middle_line_change() -> None:
    python_snapshot = render_python_tui_middle_line_change()
    ts_snapshot = run_node_reference("tui-middle-line-change")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_tui_clear_then_content() -> None:
    python_snapshot = render_python_tui_clear_then_content()
    ts_snapshot = run_node_reference("tui-clear-then-content")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]


def test_parity_tui_style_reset() -> None:
    python_snapshot = render_python_tui_style_reset()
    ts_snapshot = run_node_reference("tui-style-reset")
    assert python_snapshot["viewport"] == ts_snapshot["viewport"]
    assert python_snapshot["italic"] == ts_snapshot["italic"]



from __future__ import annotations

import asyncio
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Generic, Literal, Protocol, TypeAlias, TypeVar

from ..ai.event_stream import AssistantMessageEventStream
from ..ai.types import (
    AssistantMessage,
    AssistantMessageEvent,
    Context,
    ImageContent,
    Message,
    Model,
    SimpleStreamOptions,
    TextContent,
    ToolCall,
    ToolResultMessage,
)

ThinkingLevel: TypeAlias = Literal["off", "minimal", "low", "medium", "high", "xhigh"]
ToolExecutionMode: TypeAlias = Literal["sequential", "parallel"]


class AgentMessageLike(Protocol):
    """Structural app-level message protocol.

    Any object with a ``role`` can be stored in agent state and emitted via
    agent events. That does not make it LLM-compatible by itself; callers must
    provide ``convert_to_llm`` when custom messages should reach the model.
    """

    role: str


# AgentMessage is intentionally broader than ai.Message so higher layers can
# keep custom app messages in state. Only convert_to_llm decides which of these
# messages are filtered out, passed through, or expanded into ai.Message values.
AgentMessage: TypeAlias = Message | AgentMessageLike
AgentToolCall: TypeAlias = ToolCall

TDetails = TypeVar("TDetails")
TParameters = TypeVar("TParameters")


def _default_model() -> Model:
    from ..ai.local_env import load_local_env
    from ..ai.models import get_model

    load_local_env()
    return get_model("openai", "gpt-4o-mini")


@dataclass(slots=True)
class AgentToolResult(Generic[TDetails]):
    content: list[TextContent | ImageContent]
    details: TDetails


AgentToolUpdateCallback: TypeAlias = Callable[[AgentToolResult[Any]], None]


@dataclass(slots=True)
class AgentTool(Generic[TParameters, TDetails]):
    name: str
    label: str
    description: str
    parameters: dict[str, Any]
    execute: Callable[
        [str, TParameters, asyncio.Event | None, AgentToolUpdateCallback | None],
        AgentToolResult[TDetails] | Awaitable[AgentToolResult[TDetails]],
    ]


@dataclass(slots=True)
class AgentContext:
    system_prompt: str = ""
    messages: list[AgentMessage] = field(default_factory=list)
    tools: list[AgentTool[Any, Any]] | None = None


@dataclass(slots=True)
class BeforeToolCallResult:
    block: bool = False
    reason: str | None = None


@dataclass(slots=True)
class AfterToolCallResult:
    content: list[TextContent | ImageContent] | None = None
    details: Any | None = None
    is_error: bool | None = None


@dataclass(slots=True)
class BeforeToolCallContext:
    assistant_message: AssistantMessage
    tool_call: AgentToolCall
    args: Any
    context: AgentContext


@dataclass(slots=True)
class AfterToolCallContext:
    assistant_message: AssistantMessage
    tool_call: AgentToolCall
    args: Any
    result: AgentToolResult[Any]
    is_error: bool
    context: AgentContext


@dataclass(slots=True)
class AgentState:
    system_prompt: str = ""
    model: Model = field(default_factory=_default_model)
    thinking_level: ThinkingLevel = "off"
    tools: list[AgentTool[Any, Any]] = field(default_factory=list)
    messages: list[AgentMessage] = field(default_factory=list)
    is_streaming: bool = False
    stream_message: AgentMessage | None = None
    pending_tool_calls: set[str] = field(default_factory=set)
    error: str | None = None


@dataclass(slots=True, kw_only=True)
class AgentLoopConfig:
    """Low-level agent loop configuration.

    ``convert_to_llm`` receives the full AgentMessage list before each model
    call. It is a batch transform on purpose: one app message may map to zero,
    one, or many ai.Message values depending on the caller's needs.
    """

    model: Model
    # Custom AgentMessageLike values stay in state unless this function maps
    # them to standard ai.Message objects. The default converter is conservative
    # and only passes through built-in ai message dataclasses.
    convert_to_llm: Callable[[list[AgentMessage]], list[Message] | Awaitable[list[Message]]]
    transform_context: Callable[
        [list[AgentMessage], asyncio.Event | None],
        Awaitable[list[AgentMessage]] | list[AgentMessage],
    ] | None = None
    get_api_key: Callable[[str], str | Awaitable[str | None] | None] | None = None
    get_steering_messages: Callable[[], list[AgentMessage] | Awaitable[list[AgentMessage]]] | None = None
    get_follow_up_messages: Callable[[], list[AgentMessage] | Awaitable[list[AgentMessage]]] | None = None
    tool_execution: ToolExecutionMode = "parallel"
    before_tool_call: Callable[
        [BeforeToolCallContext, asyncio.Event | None],
        BeforeToolCallResult | Awaitable[BeforeToolCallResult | None] | None,
    ] | None = None
    after_tool_call: Callable[
        [AfterToolCallContext, asyncio.Event | None],
        AfterToolCallResult | Awaitable[AfterToolCallResult | None] | None,
    ] | None = None
    stream_options: SimpleStreamOptions | dict[str, Any] | None = None


@dataclass(slots=True, kw_only=True)
class AgentOptions:
    initial_state: dict[str, Any] | None = None
    # Required when callers want custom AgentMessageLike values to affect model
    # input. This is a batch transform, not a per-message encoder.
    convert_to_llm: Callable[[list[AgentMessage]], list[Message] | Awaitable[list[Message]]] | None = None
    transform_context: Callable[
        [list[AgentMessage], asyncio.Event | None],
        Awaitable[list[AgentMessage]] | list[AgentMessage],
    ] | None = None
    steering_mode: Literal["all", "one-at-a-time"] = "one-at-a-time"
    follow_up_mode: Literal["all", "one-at-a-time"] = "one-at-a-time"
    stream_fn: Callable[
        [Model, Context, SimpleStreamOptions | dict[str, Any] | None],
        AssistantMessageEventStream | Awaitable[AssistantMessageEventStream],
    ] | None = None
    get_api_key: Callable[[str], str | Awaitable[str | None] | None] | None = None
    stream_options: SimpleStreamOptions | dict[str, Any] | None = None
    tool_execution: ToolExecutionMode = "parallel"
    before_tool_call: Callable[
        [BeforeToolCallContext, asyncio.Event | None],
        BeforeToolCallResult | Awaitable[BeforeToolCallResult | None] | None,
    ] | None = None
    after_tool_call: Callable[
        [AfterToolCallContext, asyncio.Event | None],
        AfterToolCallResult | Awaitable[AfterToolCallResult | None] | None,
    ] | None = None


@dataclass(slots=True)
class AgentStartEvent:
    type: Literal["agent_start"] = "agent_start"


@dataclass(slots=True)
class AgentEndEvent:
    messages: list[AgentMessage]
    type: Literal["agent_end"] = "agent_end"


@dataclass(slots=True)
class TurnStartEvent:
    type: Literal["turn_start"] = "turn_start"


@dataclass(slots=True)
class TurnEndEvent:
    message: AgentMessage
    tool_results: list[ToolResultMessage]
    type: Literal["turn_end"] = "turn_end"


@dataclass(slots=True)
class MessageStartEvent:
    message: AgentMessage
    type: Literal["message_start"] = "message_start"


@dataclass(slots=True)
class MessageUpdateEvent:
    message: AgentMessage
    assistant_message_event: AssistantMessageEvent
    type: Literal["message_update"] = "message_update"


@dataclass(slots=True)
class MessageEndEvent:
    message: AgentMessage
    type: Literal["message_end"] = "message_end"


@dataclass(slots=True)
class ToolExecutionStartEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    type: Literal["tool_execution_start"] = "tool_execution_start"


@dataclass(slots=True)
class ToolExecutionUpdateEvent:
    tool_call_id: str
    tool_name: str
    args: Any
    partial_result: Any
    type: Literal["tool_execution_update"] = "tool_execution_update"


@dataclass(slots=True)
class ToolExecutionEndEvent:
    tool_call_id: str
    tool_name: str
    result: Any
    is_error: bool
    type: Literal["tool_execution_end"] = "tool_execution_end"


AgentEvent: TypeAlias = (
    AgentStartEvent
    | AgentEndEvent
    | TurnStartEvent
    | TurnEndEvent
    | MessageStartEvent
    | MessageUpdateEvent
    | MessageEndEvent
    | ToolExecutionStartEvent
    | ToolExecutionUpdateEvent
    | ToolExecutionEndEvent
)
